"""
YAML session loader
~~~~~~~~~~~~~~~~~~~

Using this code, we can create a hardware map and associated data.
"""

__all__ = [
    "HWMCSVConstructor",
    "HWMConstructor",
    "IncludedYAMLValue",
    "YAMLLoader",
    "load_from_database",
    "load_session",
    "read_csv",
    "set_yaml_loader_class",
]

import yaml
import json
import csv
import os
import time
import mimetypes
import logging
from . import hardware_map
import re


def read_csv(filename):
    with open(filename, "r") as f:
        # First pass: look for fields containing spaces and ensure they're
        # quoted.
        for l, line in enumerate(f):
            if line.startswith("#"):
                continue
            for field in re.findall(r"([^\t]* [^\t]*)", line.rstrip("\n")):
                if field[0] != '"' or field[-1] != '"':
                    raise ValueError(
                        f"File {filename} (line {l+1}). Unescaped space in field: "
                        f"'{field}'. Spaces are only legal in fields that begin and "
                        f"end with double-quote blocks. Fields should be "
                        f"tab-delimited."
                    )

        # Second pass: Actually read the file
        f.seek(0)
        dr = csv.DictReader(
            [row for row in f if not row.startswith("#")], dialect="excel-tab"
        )
        dr = list(dr)

        # Third pass: remove empty entries
        for row in dr:
            for k, v in list(row.items()):
                if v == "" or v is None:
                    row.pop(k)

    return dr


class HWMCSVConstructor:
    """When parsing a !tagged CSV filename, generate instances of 'cls'.

    The arguments to 'cls' are taken from columns in the CSV, after applying
    any transforms. For a description of transforms, see the
    HWMConstructor DocStrings.
    """

    def __init__(self, constructor, *transforms):
        self._constructor = constructor
        self._transforms = transforms

    def __call__(self, loader, node):
        fn = os.path.join(os.path.dirname(loader.name), node.value)

        classes = []

        dr = read_csv(fn)
        for m in dr:
            for t in self._transforms:
                t(loader, m)
            c = self._constructor(loader)(**m)
            classes.append(c)

        if hasattr(loader, "hwm"):
            loader.hwm.add_all(classes)

        return classes


class HWMConstructor:
    """When parsing a !tagged YAML dictionary, generate an instance of 'cls'.

    The arguments to 'cls' are taken from the YAML dictionary, after applying
    any transforms. Transforms are selected by keys in the YAML dictionary, and
    are functions of the form:

        def transform(x):
            x['foo'] = x['foo'] + 1

    This transform cause the YAML

        !some_tag { foo: 1 }

    to be instantiated as 'cls(foo=2)'. We use this to work around impedance
    mismatches between sensibly serialized HWM and the ORM.
    """

    def __init__(self, constructor, *transforms):
        self._constructor = constructor
        self._transforms = transforms

    def __call__(self, loader, node):
        # 'deep=True' is crucial and pretty much undocumented.
        m = loader.construct_mapping(node, deep=True)

        # Apply transformations, if any.
        for t in self._transforms:
            t(loader, m)

        c = self._constructor(loader)(**m)

        if hasattr(loader, "hwm"):
            loader.hwm.add(c)

        return c


class IncludedYAMLValue:
    """Container for YAML data coming from an !include snippet.

    There are two intended uses for this class:

    1. To allow structured, human-written HardwareMaps, by allowing one
       file to include another. In this mode, the save() method should
       not be used since it strips formatting from the file (e.g.
       comments).

    2. To allow bits of the hardware map to be updated by software,
       using the save() method. These files don't contain comments and
       can be reformatted and rewritten. The save() method does this.
    """

    def __new__(cls, filename, value):
        """Produce an IncludedYAMLValue class with the correct inheritence.

        Note there are two IncludedYAMLValue classes; the one above is
        user-facing, but a thin wrapper around the one below.
        """

        class IncludedYAMLValue(value.__class__):
            __doc__ = cls.__doc__  # preserve DocStrings

            def __repr__(self):
                return "IncludedYAMLValue(%r)" % value

            @property
            def filename(self):
                return filename

            def save(self):
                """Save the data contained in this file back to JSON/YAML.

                Note that comments (and other unparsed information) are
                clobbered. This method is not suitable for round-tripping
                complete HWMs.
                """
                with open(self.filename, "w") as f:

                    mimetype = mimetypes.guess_type(self.filename[0])

                    if mimetype == "application/json":
                        json.dump(value.__class__(self), f, indent=4)
                    else:
                        yaml.dump(
                            value.__class__(self), f, indent=4, default_flow_style=False
                        )

        # Instantiate and return an IncludedYAMLValue with the right value.
        return IncludedYAMLValue(value)


def yaml_include_constructor(loader, node):
    """!include tag for YAML.

    Since YAML is a superset of JSON, this tag works for JSON data too.
    """

    try:
        o = loader.construct_mapping(node, deep=True)
        retval = IncludedYAMLValue(node.value, o)

    except yaml.loader.ConstructorError:

        fn = os.path.join(os.path.dirname(loader.name), node.value)
        with open(fn) as f:
            o = yaml.safe_load(f)
        retval = IncludedYAMLValue(fn, o)

    if "save_database" in o:
        dbfile = o["save_database"]
        if not os.path.isabs(dbfile):
            dbfile = os.path.join(os.path.dirname(loader.name), dbfile)
        loader.database_path = dbfile
    # Instantiate an instance of the wrapper class with the correct value and
    # return it.
    return retval


def hwm_constructor(loader, node):
    """A YAML !HardwareMap constructor for core.HardwareMap objects.

    The node annotated with this tag must be an array of objects to be added.
    """

    # Because parts of the YAML script can query the HWM, we need to populate
    # the HWM greedily rather than building a list of objects and adding them
    # all at the end. To do so, we embed a HWM in the loader and add elements
    # as they're built.
    loader.hwm = hardware_map.HardwareMap()

    # 'deep=True' is crucial and pretty much undocumented.
    loader.construct_sequence(node, deep=True)

    return loader.hwm


def hwm_lookup_constructor(loader, node):
    """Permit YAML to make direct HWM accesses."""

    s = loader.construct_sequence(node)
    obj = s.pop(0)

    for lookup in s:
        if not isinstance(lookup, dict) or len(lookup) != 1:
            raise yaml.YAMLError(
                "HWM lookups require single key:value pairs!" "Got %r" % lookup
            )

        key = lookup.keys()[0]
        value = lookup[key]

        # Do an ORM lookup.
        try:
            obj = getattr(obj, key)[value]
        except KeyError:
            raise KeyError("%r has no %s %s" % (obj, key, value))

    return obj


class YAMLLoader(yaml.SafeLoader):

    # We allow some hooks to be queued when the document
    # is finished loading.
    __finalize_hooks = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Plumbing
        self.add_constructor("!include", yaml_include_constructor)
        self.add_constructor("!HardwareMap", hwm_constructor)
        self.add_constructor("!HWMLookup", hwm_lookup_constructor)

    def register_finalization_hook(self, hook):
        self.__finalize_hooks.append(hook)

    def construct_document(self, *args, **kwargs):
        """
        Make sure SQL database is up to date when the loader is finished.
        """
        t1 = time.time()
        data = super().construct_document(*args, **kwargs)
        # Write any pending transactions to the database
        if hasattr(self, "hwm"):
            self.hwm.commit()
            # Store constructed HWM to disk
            if hasattr(self, "database_path"):
                self.hwm.dump(self.database_path, backup=True)

        # Fire any callbacks
        for hook in self.__finalize_hooks:
            hook(getattr(self, "hwm", None))

        t2 = time.time()
        if hasattr(self, "hwm"):
            logging.info("Hardware map loaded in %.2f seconds" % (t2 - t1))

        return data


def set_yaml_loader_class(cls):
    """Override the YAMLLoader class used to create sessions."""

    global __yaml_loader_class
    __yaml_loader_class = cls


__yaml_loader_class = YAMLLoader


def load_session(stream, store=True):
    """Load a YAML document into a Session object."""
    y = yaml.load(stream, Loader=__yaml_loader_class)
    if hasattr(stream, "name") and "hardware_map" in y:
        y["hwm_dir"] = os.path.abspath(stream.name)
    if store:
        set_session(y)
    return y


def set_session(session):
    """Store a Session object somewhere it's globally accessible."""
    global __session_handle
    __session_handle = session


def get_session():
    """Retrieve the session stored via set_session()"""
    return __session_handle


__session_handle = None


def load_from_database(path):
    """
    To use within yaml hwm:
    output_database : !include
        save_database : '/tmp/rfmux.db'
    """

    hwm = hardware_map.HardwareMap("sqlite:///{path}")

    # preload attributes for tab-completion in interactive sessions
    hwm.cache_attributes()

    return hwm


# vim: sts=4 ts=4 sw=4 tw=80 smarttab expandtab



''' Class to extract parameters from the existing dialog and compare it with the default test'''

import inspect
import ast
from importlib import import_module

class ParamKeyExtractor:
    """Extracts parameter keys used inside `<class>.get_parameters()`."""

    PARAM_CONTAINERS = {"params", "params_dict"}

    def __init__(self, module_path: str, class_name: str):
        self.module_path = module_path
        self.class_name = class_name
        self.tree = self._parse_module_ast()

    # ----------------------------
    # Parsing + Class/Method Lookup
    # ----------------------------

    def _parse_module_ast(self) -> ast.Module:
        try:
            module = import_module(self.module_path)
            source = inspect.getsource(module)
            return ast.parse(source)
        except Exception as exc:  # pragma: no cover
            raise RuntimeError(
                f"Unable to parse {self.module_path} for parameter key validation: {exc}"
            ) from exc

    def _find_class(self) -> ast.ClassDef | None:
        for node in self.tree.body:
            if isinstance(node, ast.ClassDef) and node.name == self.class_name:
                return node
        return None

    def _find_method(self, class_node: ast.ClassDef) -> ast.FunctionDef | None:
        for node in class_node.body:
            if isinstance(node, ast.FunctionDef) and node.name == "get_parameters":
                return node
        return None

    # ----------------------------
    # Extraction Rules
    # ----------------------------

    def _extract_keys_from_dict(self, node: ast.Dict) -> set[str]:
        return {
            key.value
            for key in node.keys
            if isinstance(key, ast.Constant) and isinstance(key.value, str)
        }

    def _extract_keys_from_assignment(self, node: ast.Assign) -> set[str]:
        """Handles patterns like: params.foo = value"""
        keys = set()
        for target in node.targets:
            if (
                isinstance(target, ast.Attribute)
                and isinstance(target.value, ast.Name)
                and target.value.id in self.PARAM_CONTAINERS
            ):
                keys.add(target.attr)
        return keys

    def _extract_keys_from_subscript(self, node: ast.Subscript) -> set[str]:
        """Handles patterns like: params['foo']"""
        if isinstance(node.value, ast.Name) and node.value.id in self.PARAM_CONTAINERS:
            slice_node = node.slice
            # Python <3.9 compatibility
            if isinstance(slice_node, ast.Index):
                slice_node = slice_node.value
            if isinstance(slice_node, ast.Constant) and isinstance(slice_node.value, str):
                return {slice_node.value}
        return set()


    # ----------------------------
    # Main Extraction
    # ----------------------------

    def extract(self) -> set[str]:
        """Main public method: returns all parameter keys found."""
        class_node = self._find_class()
        if not class_node:
            return set()

        method_node = self._find_method(class_node)
        if not method_node:
            return set()

        keys: set[str] = set()

        for node in ast.walk(method_node):
            if isinstance(node, ast.Dict):
                keys.update(self._extract_keys_from_dict(node))
            elif isinstance(node, ast.Assign):
                keys.update(self._extract_keys_from_assignment(node))
            elif isinstance(node, ast.Subscript):
                keys.update(self._extract_keys_from_subscript(node))

        return keys



"""
Mock CRS Device - UDP Streaming Logic.
Manages UDP packet generation and streaming for MockCRS.
"""
import asyncio
import socket
import struct
import time
import threading
import signal
import atexit
from datetime import datetime
import numpy as np
import platform

# Import ReadoutPacket and related constants from streamer
from ..streamer import (
    ReadoutPacket, Timestamp, TimestampSource,
    STREAMER_MAGIC,
    SHORT_PACKET_VERSION, LONG_PACKET_VERSION,
    SHORT_PACKET_CHANNELS, LONG_PACKET_CHANNELS,
    SS_PER_SECOND,
    STREAMER_PORT, # Default port for streaming
)


# Global registry to track all active UDP streamers for cleanup
_active_streamers = []
_cleanup_registered = False

def _emergency_cleanup():
    """Emergency cleanup function called on program exit."""
    global _active_streamers
    if _active_streamers:
        for streamer in _active_streamers[:]:  # Copy list to avoid modification during iteration
            try:
                streamer.emergency_stop()
            except Exception as e:
                print(f"[UDP] Error during emergency cleanup: {e}")

def _register_global_cleanup():
    """Register global cleanup handlers (called once)."""
    global _cleanup_registered
    if not _cleanup_registered:
        atexit.register(_emergency_cleanup)
        # Register signal handlers for graceful shutdown
        for sig in [signal.SIGTERM, signal.SIGINT]:
            try:
                signal.signal(sig, lambda signum, frame: _emergency_cleanup())
            except (ValueError, OSError):
                pass  # Signal handling may not be available in all contexts
        _cleanup_registered = True

class MockCRSUDPStreamer(threading.Thread):
    """Streams UDP packets with S21 data based on MockCRS state"""
    
    def __init__(self, mock_crs, host='239.192.0.2', port=STREAMER_PORT, modules_to_stream=None, use_multicast=True):
        super().__init__(daemon=True)
        self.mock_crs = mock_crs # Reference to MockCRS instance
        self.host = host
        self.port = port
        self.use_multicast = use_multicast
        
        # Configure for multicast or unicast
        if use_multicast:
            # Use multicast for better hardware emulation
            self.multicast_group = host if host.startswith('239.') else '239.192.0.2'
            self.multicast_port = port
            print(f"[UDP] MockCRSUDPStreamer initialized - multicast to {self.multicast_group}:{self.multicast_port}")
        else:
            # Fall back to unicast for specific testing scenarios
            self.unicast_host = host if host != '239.192.0.2' else '127.0.0.1'
            self.unicast_port = port
            print(f"[UDP] MockCRSUDPStreamer initialized - unicast to {self.unicast_host}:{self.unicast_port}")
        
        # If modules_to_stream is specified, only stream those modules
        # Otherwise, we'll determine which modules to stream based on configured channels
        self.modules_to_stream = modules_to_stream  # Can be None or a list like [1] or [1,2]
        
        self.socket = None
        self.running = False # Controlled by start/stop
        self.seq_counters = {m: 0 for m in range(1, 5)}  # Per-module sequence numbers (1-4)
        self.packets_sent = 0
        
        # Track total elapsed time for continuous timestamps across decimation changes
        self.total_elapsed_time = {m: 0.0 for m in range(1, 5)}  # Total time elapsed per module
        self.last_decimation = None

        self.timestamp_stream = None
        
        # Register for global cleanup
        global _active_streamers
        _active_streamers.append(self)
        _register_global_cleanup()
        
    def _init_socket(self):
        """Initialize the UDP socket for multicast or unicast."""
        if self.socket is None:
            if self.use_multicast:
                # Create socket for multicast
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

                
                # Enable multicast loopback so local processes can receive
                self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1)
                
                # Set socket send buffer size for performance
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 4_000_000)

                if platform.system() == "Windows":
                    try:
                        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
                        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton("127.0.0.1"))
                    except:
                        print("Issue setting up socket on windows")
                    
                
                # Bind multicast to loopback interface (lo) for local testing
                # Using if_nametoindex to get the interface index for 'lo'  
                else:
                    for iface in ("lo", "lo0"): 
                        # lo is for Linux and lo0 is for Mac 
                        try:
                            lo_index = socket.if_nametoindex(iface)
                            # Use IP_MULTICAST_IF with the loopback address
                            # Note: IP_MULTICAST_IF expects an IP address, not an interface index
                            # For interface index, we'd use IPV6_MULTICAST_IF, but we're using IPv4
                            # So we stick with the loopback IP address
                            self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton('127.0.0.1'))
                            print(f"[UDP] Multicast bound to loopback interface ({iface}, index {lo_index})")
                            break
                        except OSError:
                            continue
                    else:
                        # Fallback if 'lo or lo0' interface not found (shouldn't happen on Linux or Mac)
                        print("[UDP] Warning: Could not find 'lo' or 'lo0' interface")
                        print("[UDP] Warning: The packets are now being launched on your network, rather than on the loopback interface")
                        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton('0.0.0.0'))
                    
                    print(f"[UDP] Multicast socket initialized for {self.multicast_group}:{self.multicast_port}")
            else:
                # Create socket for unicast
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 4_000_000)
                print(f"[UDP] Unicast socket initialized")
    
    def _cleanup_socket(self):
        """Clean up the UDP socket."""
        if self.socket:
            try:
                self.socket.close()
            except Exception as e:
                print(f"[UDP] Error closing socket: {e}")
            finally:
                self.socket = None
        
    def _get_configured_modules(self):
        """Determine which modules have configured channels"""
        configured_modules = set()
        
        # Check frequencies, amplitudes, and phases dictionaries
        for (module, channel) in self.mock_crs._frequencies.keys():
            configured_modules.add(module)
        for (module, channel) in self.mock_crs._amplitudes.keys():
            configured_modules.add(module)
        for (module, channel) in self.mock_crs._phases.keys():
            configured_modules.add(module)
        
        # Convert to sorted list and ensure it's within valid range (1-4)
        modules = sorted([m for m in configured_modules if 1 <= m <= 4])
        
        if not modules:
            # If no modules configured, default to module 1
            modules = [1]
            
        return modules
    
    def stop(self):
        """Stop the streaming thread gracefully."""
        print("[UDP] Stopping UDP streamer...")
        self.running = False
        
    def emergency_stop(self):
        """Emergency stop - force shutdown immediately."""
        print(f"[UDP] Emergency stop for streamer {id(self)}")
        self.running = False
        self._cleanup_socket()
        # Remove from global registry
        global _active_streamers
        if self in _active_streamers:
            _active_streamers.remove(self)
    
    def run(self):
        """Stream UDP packets at the configured sample rate with proper cleanup."""
        self.running = True
        
        # Initialize start time for deterministic timestamps
        self.start_datetime = datetime.now()
        
        last_decimation = None
        
        try:
            # Initialize socket
            self._init_socket()
            
            while self.running:
                try:
                    start_time_loop = time.perf_counter()
                    
                    # Get current decimation (dynamically check each iteration)
                    dec = self.mock_crs._fir_stage
                    if dec is None:
                        dec = 6  # Default only if None, not if 0
                    sample_rate = 625e6 / (256 * 64 * (2**dec))
                    frame_time = 1.0 / sample_rate  # Time between frames
                    
                    # Log if decimation changed
                    if dec != last_decimation:
                        # Before changing decimation, save the elapsed time for each module
                        if last_decimation is not None:
                            old_sample_rate = 625e6 / (256 * 64 * (2**last_decimation))
                            for m in self.seq_counters:
                                # Add the time elapsed with the old decimation rate
                                self.total_elapsed_time[m] += self.seq_counters[m] / old_sample_rate
                        
                        print(f"[UDP] Decimation changed to stage {dec}, streaming at {sample_rate:.1f} Hz")
                        last_decimation = dec
                        self.last_decimation = dec
                        
                    # Determine which modules to stream
                    if self.modules_to_stream:
                        # Use explicitly specified modules
                        modules_to_process = self.modules_to_stream
                    else:
                        # Auto-detect: only stream modules that have configured channels
                        modules_to_process = self._get_configured_modules()
                    
                    # Generate and send packets for ALL active modules quickly
                    # Each module gets its own packet at the same time instant
                    for module_num in modules_to_process:
                        if not self.running:  # Check if we should stop
                            break
                            
                        try:
                            # Generate and send packet
                            s_time = time.perf_counter()
                            packet_bytes = self.generate_packet_for_module(module_num, self.seq_counters[module_num], dec)
                            
                            if self.use_multicast:
                                if self.socket:
                                    try:
                                        bytes_sent = self.socket.sendto(packet_bytes, (self.multicast_group, self.multicast_port))
                                    except OSError as e:
                                        if e.errno == 10051: ## unreachable network on Windows, corner case on first start and for tests
                                            continue
                                        else:
                                            raise
                                else:
                                    raise RuntimeError("Socket not initialized")
                            else:
                                if self.socket:
                                    bytes_sent = self.socket.sendto(packet_bytes, (self.unicast_host, self.unicast_port))
                                else:
                                    raise RuntimeError("Socket not initialized")
                            
                            self.seq_counters[module_num] += 1
                            self.packets_sent += 1
                            
                        except Exception as e:
                            if self.running:  # Only log if we're still supposed to be running
                                print(f"[UDP] ERROR generating/sending packet for module {module_num}: {e}")
                                import traceback
                                traceback.print_exc()
                    
                    # Sleep to maintain sample rate
                    if self.running:
                        elapsed_loop = time.perf_counter() - start_time_loop
                        sleep_duration = frame_time - elapsed_loop
                        
                        if sleep_duration > 0:
                            # Simple sleep approach - let the OS handle timing
                            time.sleep(sleep_duration)
                
                except KeyboardInterrupt:
                    print("[UDP] Received KeyboardInterrupt, stopping...")
                    break
                except Exception as e:
                    if self.running:
                        print(f"[UDP] Unexpected error in main loop: {e}")
                        import traceback
                        traceback.print_exc()
                        time.sleep(0.1)  # Brief pause before retrying
            
        except Exception as e:
            print(f"[UDP] Fatal error in UDP streamer: {e}")
            import traceback
            traceback.print_exc()
        
        finally:
            # ALWAYS clean up, regardless of how we exit
            print(f"[UDP] UDP streaming thread stopped. Total packets sent: {self.packets_sent}")
            self._cleanup_socket()
            
            # Remove from global registry
            global _active_streamers
            if self in _active_streamers:
                _active_streamers.remove(self)
    
    def generate_packet_for_module(self, module_num, seq, dec):
        """Generate a ReadoutPacket for a specific module with coupled channels."""

        # Detailed timing
        timing_start = time.perf_counter()
        
        if self.mock_crs._short_packets:
            num_channels = SHORT_PACKET_CHANNELS
            version = SHORT_PACKET_VERSION
        else:
            num_channels = LONG_PACKET_CHANNELS
            version = LONG_PACKET_VERSION

        timing_array_create = time.perf_counter()

        # Pre-calculate constants from unified configuration (SoT)
        cfg = getattr(self.mock_crs, "physics_config", {}) or {}
        scale_factor = cfg.get("scale_factor", 2**21)
        full_scale = scale_factor * 256. # Full scale in normalized units, scaled up by 256 to account for 24-bit scalar packing into 32-bit packets
        noise_level = cfg.get("udp_noise_level", 10.0)  # Base noise level

        # Noise is in normalized units
        noise_i = np.random.normal(0, noise_level, num_channels)
        noise_q = np.random.normal(0, noise_level, num_channels)
        channel_samples = noise_i + 1j * noise_q

        timing_noise = time.perf_counter()

        # Use module-wide coupled calculation if available
        if hasattr(self.mock_crs._resonator_model, 'calculate_module_response_coupled'):
            # Get sample rate for time-varying signals
            dec = dec
            if dec is None:
                dec = 6  # Default only if None, not if 0
            sample_rate = 625e6 / 256 / 64 / (2**dec)
            
            # Calculate time for this packet (based on sequence number)
            # Each packet represents one sample in time
            t = seq / sample_rate
            
            # Update QP densities based on current time (for pulse events)
            if hasattr(self.mock_crs._resonator_model, 'update_qp_densities_for_time'):
                self.mock_crs._resonator_model.update_qp_densities_for_time(t)
            
            # Count configured channels for debugging
            num_configured = 0
            for (mod, ch) in self.mock_crs._frequencies.keys():
                if mod == module_num:
                    num_configured += 1
            
            # Get all channel responses at once (includes coupling effects and time-varying beats)
            # Now using the unified method with start_time parameter
            channel_responses = self.mock_crs._resonator_model.calculate_module_response_coupled(
                module_num, 
                num_samples=1,  # Single sample per packet
                sample_rate=sample_rate,
                start_time=t   # Pass the current time for this packet
            )
            timing_physics = time.perf_counter()
            
            # Process each channel's response
            for ch_num_1, signal in channel_responses.items():
                ch_idx_0 = ch_num_1 - 1  # Convert to 0-based index
                
                # Scale signal from normalized units to ADC counts and add to noise
                # signal is in normalized units, full_scale converts to ADC counts
                channel_samples[ch_idx_0] += signal * full_scale
                
            timing_fill_array = time.perf_counter()
            non_zero_channels_count = len(channel_responses)
        else:
            print("ERROR WITH COUPLING CODE - this shouldn't happen with updated resonator model")
            non_zero_channels_count = 0
            timing_physics = timing_fill_array = time.perf_counter()
        
        # Calculate deterministic timestamp based on sequence number and sampling rate
        # Get the decimation stage and calculate sample rate
        dec = dec
        if dec is None:
            dec = 6  # Default only if None, not if 0
        sample_rate = 625e6 / 256 / 64 / (2**dec)
        
        # Calculate total elapsed time including previous decimation periods
        # This ensures continuous timestamps even when decimation changes
        current_period_elapsed = seq / sample_rate
        total_elapsed_seconds = self.total_elapsed_time.get(module_num, 0.0) + current_period_elapsed
        
        # Add elapsed time to start datetime to get packet timestamp
        # Use the start_datetime that was set when streaming began
        if hasattr(self, 'start_datetime') and self.start_datetime:
            base_time = self.start_datetime
        else:
            # Fallback if start_datetime not set (shouldn't happen)
            base_time = datetime.now()
            self.start_datetime = base_time
        
        # Calculate the timestamp for this packet using total elapsed time
        from datetime import timedelta
        packet_datetime = base_time + timedelta(seconds=total_elapsed_seconds)
        
        # Convert to timestamp fields
        ts = Timestamp(
            y=int(packet_datetime.year % 100),
            d=int(packet_datetime.timetuple().tm_yday),
            h=int(packet_datetime.hour),
            m=int(packet_datetime.minute),
            s=int(packet_datetime.second),
            ss=int(packet_datetime.microsecond * SS_PER_SECOND / 1e6), # Scaled sub-seconds
            c=0, # Carrier phase, typically 0 for mock
            sbs=0, # Sub-block sequence, typically 0 for mock
            source=TimestampSource.TEST, # Mock data source
            recent=True
        )

        timing_timestamp = time.perf_counter()
        packet = ReadoutPacket(
            magic=STREAMER_MAGIC,
            version=version,
            serial=int(self.mock_crs._serial) if self.mock_crs._serial and self.mock_crs._serial.isdigit() else 0,
            num_modules=1, # Packet is for one module's data
            flags=0,
            fir_stage=dec,
            module=module_num-1,  # ReadoutPacket module is 0-indexed
            seq=seq)

        # Clip and assign complex samples to packet
        packet.samples = (np.clip(channel_samples.real, -8388608, 8388607) +
                1j*np.clip(channel_samples.imag, -8388608, 8388607)).tolist()
        packet.ts = ts

        timing_packet_create = time.perf_counter()

        packet_bytes = bytes(packet)
        self.mock_crs._last_timestamp = ts
        timing_serialize = time.perf_counter()
        
        # Removed detailed timing logs - no longer needed
        
        return packet_bytes


class MockUDPManager:
    """Manages the lifecycle of the MockCRSUDPStreamer thread with proper cleanup."""
    def __init__(self, mock_crs):
        self.mock_crs = mock_crs
        self.udp_thread = None
        self._udp_streaming_active = False # Internal flag

    async def start_udp_streaming(self, host='239.192.0.2', port=STREAMER_PORT, use_multicast=True):
        """Start UDP streaming with proper error handling.
        
        Args:
            host: Multicast group address (default: '239.192.0.2') or unicast IP
            port: UDP port number (default: STREAMER_PORT)
            use_multicast: If True, use multicast; if False, use unicast
        """
        try:
            if not self._udp_streaming_active:
                self._udp_streaming_active = True
                
                # Create the streamer with multicast or unicast configuration
                self.udp_thread = MockCRSUDPStreamer(
                    self.mock_crs, 
                    host=host, 
                    port=port,
                    use_multicast=use_multicast
                )
                self.udp_thread.start()
                await asyncio.sleep(0.1) # Give thread time to start
                
                mode = "multicast" if use_multicast else "unicast"
                print(f"[Manager] UDP Streaming started ({mode}) to {host}:{port}")
                return True
            else:
                print("[Manager] UDP Streaming already active.")
                return False # Already active
        except Exception as e:
            print(f"[Manager] Error starting UDP streaming: {e}")
            self._udp_streaming_active = False
            if self.udp_thread:
                self.udp_thread.emergency_stop()
                self.udp_thread = None
            raise
    
    async def stop_udp_streaming(self):
        """Stop UDP streaming with proper cleanup and timeout."""
        if self._udp_streaming_active and self.udp_thread:
            try:
                self._udp_streaming_active = False
                self.udp_thread.stop() # Signal thread to stop
                
                # Wait for thread to finish with timeout
                try:
                    await asyncio.to_thread(self.udp_thread.join, timeout=2.0)
                except Exception as e:
                    print(f"[Manager] Error during thread join: {e}")
                
                if self.udp_thread.is_alive():
                    print("[Manager] Warning: UDP thread did not stop in time, forcing emergency stop")
                    self.udp_thread.emergency_stop()
                
                self.udp_thread = None
                print("[Manager] UDP Streaming stopped.")
                return True
                
            except Exception as e:
                print(f"[Manager] Error stopping UDP streaming: {e}")
                # Force cleanup in case of error
                if self.udp_thread:
                    self.udp_thread.emergency_stop()
                    self.udp_thread = None
                return False
        else:
            print("[Manager] UDP Streaming not active or thread missing.")
            return False # Not active or no thread

    def get_udp_streaming_status(self):
        return {
            "active": self._udp_streaming_active,
            "thread_alive": self.udp_thread is not None and self.udp_thread.is_alive()
        }
    
    def __del__(self):
        """Destructor to ensure cleanup."""
        if self._udp_streaming_active and self.udp_thread:
            print("[Manager] Destructor cleanup: stopping UDP streamer")
            self.udp_thread.emergency_stop()
    
    def __enter__(self):
        """Context manager entry."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit with cleanup."""
        if self._udp_streaming_active and self.udp_thread:
            print("[Manager] Context exit cleanup: stopping UDP streamer")
            self.udp_thread.emergency_stop()



#!/usr/bin/env -S PYTHONPATH=.. pytest-3 -v -W error::UserWarning

"""
Spot checks for API regressions.

This test script can be invoked in three ways:

- By itself, in your PC's current Python environment, without relying on a
  running CRS board:

      ./test_spotcheck.py

- By itself, in your PC's current Python environment, with a CRS board running
  alongside:

      ./test_spotcheck.py --serial=0024

- As part of a complete regression test environment (exercising all test
  scripts and in a variety of different Python environments), without relying
  on a running CRS board:

      ~/rfmux$ ./test.sh
"""

import rfmux
import pytest
import numpy as np

@pytest.mark.asyncio
@pytest.mark.parametrize(
    "frequency", (100e6, 2.4e9, 2.6e9, 4.9e9, 5.1e9, 7.4e9, 7.6e9, 9.9e9))
@pytest.mark.parametrize("module", range(1, 9))
async def test_nco_set_get_frequencies(crs, frequency, module):
    """
    Do NCO setters and getters work correctly?

    The purpose is to check that the get/set correctly manages higher Nyquist
    zones in the C API. Note that it doesn't guarantee the RFDC is actually
    doing the right thing - just API consistency.
    """

    zone = int(frequency / 2.5e9) + 1

    await crs.set_analog_bank(high=module > 4)

    if zone <= 2:
        # Works as expected without API warning
        await crs.set_nco_frequency(frequency, module=module)
    else:
        # Nyquist zones 3 and 4 currently emit a bogus warning from the RFDC
        # API
        with pytest.warns(UserWarning):
            await crs.set_nco_frequency(frequency, module=module)

    assert (await crs.get_nco_frequency(module=module)) == pytest.approx(
        frequency, abs=0.001
    )  # expect agreement below 1 mHz

    dac_filter = await crs.get_dac_compensation_filter(module=module)

    match(zone):
        case 1:
            assert dac_filter==crs.DAC_COMPENSATION_FILTER.MODE1
        case 2:
            assert dac_filter==crs.DAC_COMPENSATION_FILTER.MODE2
        case 3:
            assert dac_filter==crs.DAC_COMPENSATION_FILTER.MODE2
        case 4:
            assert dac_filter==crs.DAC_COMPENSATION_FILTER.MODE1
        case _:
            assert False


@pytest.mark.asyncio
async def test_set_get_decimation(crs):
    """
    Does set_decimation and get_decimation operate correctly?
    """

    # clearly this should fail - requesting absurd amounts of bandwidth
    with pytest.raises(rfmux.tuber.TuberRemoteError):
        await crs.set_decimation(stage=0, short=False, modules=[1, 2, 3, 4])

    # we should be able to stream nothing, but can't get a value from get_decimation like this
    await crs.set_decimation(stage=6, short=False, modules=[])
    assert (await crs.get_decimation()) is None

    # With short packets, we can go all the way down to FIR stage 0
    # (but only with one module at a time)
    for module in range(1, 5):
        for stage in range(0, 7):
            await crs.set_decimation(stage=stage, short=True, modules=[module])
            assert (await crs.get_decimation()) == stage

    # With longer packets, we can only do this at FIR stage 3 and up
    for module in range(1, 5):
        for stage in range(3, 7):
            await crs.set_decimation(stage=stage, short=False, modules=[module])
            assert (await crs.get_decimation()) == stage

    # be polite
    await crs.set_decimation(stage=6, short=False, modules=[1, 2, 3, 4])


@pytest.mark.asyncio
async def test_high_sampling_rate(crs):
    """
    Are samples arriving correctly ordered, even at high sampling rates?

    This verifies the (limited) reorder buffer in both py_get_samples and
    get_samples, and indirectly checks the network and streamer capacity.
    Both of these needed attention in the lead-up to v1.5.6, so this test
    checks against associated regressions.
    """

    # fastest streaming available for a single module
    await crs.set_decimation(stage=0, short=True, modules=[1])

    # py_get_samples
    x = await crs.py_get_samples(1000, channel=1, module=1, _extra_metadata=True)
    assert {*np.diff(x.seq)} == {1}

    # get_samples
    x = await crs.get_samples(1000, channel=1, module=1, _extra_metadata=True)
    assert {*np.diff(x.seq)} == {1}



@pytest.mark.asyncio
async def test_py_get_samples_long_and_short(crs):
    """
    Does py_get_samples play nice with long and short packets?
    """

    await crs.set_decimation(stage=6, short=True, modules=[1])

    # Try with "channel" specified and low
    x = await crs.py_get_samples(10, channel=1, module=1)
    assert len(x.i) == 10 and len(x.q) == 10

    # Try with "channel" specified and too high
    with pytest.raises(ValueError):
        x = await crs.py_get_samples(10, channel=129, module=1)

    # Now don't specify "channel" - should get an array of 128 channels
    x = await crs.py_get_samples(10, module=1)
    assert len(x.i) == 128 and len(x.q) == 128

    # Back to long packets
    await crs.set_decimation(stage=6, short=False, modules=[1, 2, 3, 4])

    # Try with "channel" specified and low
    x = await crs.py_get_samples(10, channel=1, module=1)
    assert len(x.i) == 10 and len(x.q) == 10

    # Now don't specify "channel" - should get an array of 1024 channels
    x = await crs.py_get_samples(10, module=1)
    assert len(x.i) == 1024 and len(x.q) == 1024


if __name__ == "__main__":
    pytest.main([__file__])



"""Panel class for network analysis results (dockable)."""
import datetime # Added for MultisweepWindow export
import pickle   # Added for MultisweepWindow export
import os 
import csv

# Imports from within the 'periscope' subpackage
from .utils import *
from .tasks import SetCableLengthSignals # Added import
# from .tasks import * # Not directly used by this class, dialogs will import what they need.

# Dialogs are now imported from .dialogs within the same package
from .dialogs import NetworkAnalysisParamsDialog, FindResonancesDialog, MultisweepDialog
from .network_analysis_export import NetworkAnalysisExportMixin

class NetworkAnalysisPanel(QtWidgets.QWidget, NetworkAnalysisExportMixin, ScreenshotMixin):
    """
    Dockable panel for displaying network analysis results with real units support.

    This panel can be wrapped in a QDockWidget for tabbed/floating display within
    the main Periscope window. All functionality from the original NetworkAnalysisWindow
    is preserved.
    
    Signals:
        data_ready: Emitted when analysis completes with full export data dict
    """
    
    # Signal for session auto-export
    data_ready = QtCore.pyqtSignal(dict)  # Full export data dictionary
    
    def __init__(self, parent=None, modules=None, dac_scales=None, dark_mode=False, is_loaded_data=False):
        super().__init__(parent)
        self.modules = modules or []
        self.data = {}  # module -> amplitude data dictionary
        self.raw_data = {}  # Store the raw IQ data for unit conversion
        self.unit_mode = "dbm"  # Default to dBm instead of counts
        self.normalize_magnitudes = False  # Add this flag to track normalization state
        self.first_setup = True  # Flag to track initial setup
        self.zoom_box_mode = True  # Default to zoom box mode ON
        self.plots = {}  # Initialize plots dictionary early
        self.original_params = {}  # Initial parameters
        self.current_params = {}   # Most recently used parameters
        self.dac_scales = dac_scales or {}  # Store DAC scales
        self.resonance_lines_mag = {} # Store resonance lines for magnitude plots {module: [line_item, ...]}
        self.resonance_lines_phase = {} # Store resonance lines for phase plots {module: [line_item, ...]}
        self.resonance_freqs = {}  # Store resonance frequencies per module
        self.add_subtract_mode = False
        self.module_cable_lengths = {} # For Requirement 2
        self.faux_resonance_legend_items_mag = {} # For Req 3
        self.faux_resonance_legend_items_phase = {} # For Req 3
        self.dark_mode = dark_mode  # Store dark mode setting
        self.is_loaded_data = is_loaded_data  # Track if this is from loaded data
        
        # Track last session export filename for overwriting
        self._last_export_filename: Optional[str] = None

        # Initialize signals for SetCableLengthTask
        self.set_cable_length_signals = SetCableLengthSignals()
        # Optionally, connect these signals to handlers for user feedback
        self.set_cable_length_signals.error.connect(self._handle_set_cable_length_error)
        
        # Setup the UI components
        self._setup_ui()
        # Set initial size only on creation
        self.resize(1000, 800)

    def _setup_ui(self):
        """Set up the user interface for the panel."""
        # Create main layout for the panel (no central widget needed for QWidget)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)  # No margins for cleaner docking
        
        # Create toolbar
        self._setup_toolbar(layout)
        
        # Create progress bars
        self._setup_progress_bars(layout)
        
        # Create plot area
        self._setup_plot_area(layout)

    def _setup_toolbar(self, layout):
        """Set up the toolbars with controls."""
        # Toolbar 1: Global Controls (Top Row)
        # Use QWidget container instead of QToolBar for compatibility with QWidget base class
        toolbar_global = QtWidgets.QWidget()
        toolbar_global_layout = QtWidgets.QHBoxLayout(toolbar_global)
        toolbar_global_layout.setContentsMargins(5, 5, 5, 5)

        # Export button
        export_btn = QtWidgets.QPushButton("💾")
        export_btn.setToolTip("Export data")
        export_btn.clicked.connect(self._export_data)
        toolbar_global_layout.addWidget(export_btn)

        # Edit Other Parameters button (renamed to Re-run Analysis)
        edit_params_btn = QtWidgets.QPushButton("Re-run analysis")
        edit_params_btn.clicked.connect(self._edit_parameters)
        toolbar_global_layout.addWidget(edit_params_btn)

        # Show/Hide resonances checkbox
        self.show_resonances_cb = QtWidgets.QCheckBox("Show Resonances")
        self.show_resonances_cb.setChecked(True)
        self.show_resonances_cb.toggled.connect(self._toggle_resonances_visible)
        toolbar_global_layout.addWidget(self.show_resonances_cb)

        # Add/Subtract mode
        self.edit_resonances_cb = QtWidgets.QCheckBox("Add/Subtract Resonances")
        self.edit_resonances_cb.setToolTip(
            "When enabled, double-click adds a resonance;\n"
            "Shift + double-click removes the nearest resonance."
        )
        self.edit_resonances_cb.toggled.connect(self._toggle_resonance_edit_mode)
        toolbar_global_layout.addWidget(self.edit_resonances_cb)

        # Add spacer to push the unit controls to the far right
        toolbar_global_layout.addStretch(1)

        # Normalize Magnitudes checkbox
        self.normalize_checkbox = QtWidgets.QCheckBox("Normalize Magnitudes")
        self.normalize_checkbox.setChecked(False)
        self.normalize_checkbox.setToolTip("Normalize all magnitude curves to their first data point")
        self.normalize_checkbox.toggled.connect(self._toggle_normalization)
        toolbar_global_layout.addWidget(self.normalize_checkbox)


        # Add unit controls
        self._setup_unit_controls(toolbar_global_layout)

        # Add zoom box mode checkbox
        self._setup_zoom_box_control(toolbar_global_layout)

        # Screenshot button
        screenshot_btn = QtWidgets.QPushButton("📷")
        screenshot_btn.setToolTip("Export a screenshot of this panel to the session folder (or choose location)")
        screenshot_btn.clicked.connect(self._export_screenshot)
        toolbar_global_layout.addWidget(screenshot_btn)
        
        layout.addWidget(toolbar_global)

        # Toolbar 2: Module-Specific Controls (Bottom Row)
        toolbar_module = QtWidgets.QWidget()
        toolbar_module_layout = QtWidgets.QHBoxLayout(toolbar_module)
        toolbar_module_layout.setContentsMargins(5, 5, 5, 5)

        # Cable length control
        self.cable_length_label = QtWidgets.QLabel("Cable Length (m):")
        self.cable_length_spin = QtWidgets.QDoubleSpinBox()
        self.cable_length_spin.setRange(0.0, 1000.0)
        self.cable_length_spin.setValue(DEFAULT_CABLE_LENGTH)
        self.cable_length_spin.setSingleStep(0.05)
        self.cable_length_spin.valueChanged.connect(self._on_cable_length_changed)
        toolbar_module_layout.addWidget(self.cable_length_label)
        toolbar_module_layout.addWidget(self.cable_length_spin)

        # Unwrap Cable Delay button
        unwrap_button = QtWidgets.QPushButton("Unwrap Cable Delay")
        unwrap_button.setToolTip("Fit phase slope, calculate cable length, and apply compensation for the active module.")
        unwrap_button.clicked.connect(self._unwrap_cable_delay_action)
        toolbar_module_layout.addWidget(unwrap_button)

        # Find Resonances button
        find_res_btn = QtWidgets.QPushButton("Find Resonances")
        find_res_btn.setToolTip("Identify resonance frequencies from the current sweep data for the active module.")
        find_res_btn.clicked.connect(self._show_find_resonances_dialog)
        toolbar_module_layout.addWidget(find_res_btn)

        # Take Multisweep button
        self.take_multisweep_btn = QtWidgets.QPushButton("Take Multisweep")
        self.take_multisweep_btn.setToolTip("Perform a multisweep using identified resonance frequencies for the active module.")
        self.take_multisweep_btn.clicked.connect(self._show_multisweep_dialog)
        self.take_multisweep_btn.setEnabled(False) # Initially disabled
        toolbar_module_layout.addWidget(self.take_multisweep_btn)
        
        toolbar_module_layout.addStretch(1)
        
        layout.addWidget(toolbar_module)

    def _setup_unit_controls(self, toolbar_layout):
        """Set up the unit selection controls and add them to the specified toolbar layout."""
        unit_group = QtWidgets.QWidget()
        unit_layout = QtWidgets.QHBoxLayout(unit_group)
        unit_layout.setContentsMargins(0, 0, 0, 0)
        unit_layout.setAlignment(Qt.AlignmentFlag.AlignRight)
        
        self.rb_counts = QtWidgets.QRadioButton("Counts")
        self.rb_dbm = QtWidgets.QRadioButton("dBm")
        self.rb_volts = QtWidgets.QRadioButton("Volts")
        self.rb_dbm.setChecked(True)
        
        unit_layout.addWidget(QtWidgets.QLabel("Units:"))
        unit_layout.addWidget(self.rb_counts)
        unit_layout.addWidget(self.rb_dbm)
        unit_layout.addWidget(self.rb_volts)
        
        # Connect signals
        self.rb_counts.toggled.connect(lambda: self._update_unit_mode("counts"))
        self.rb_dbm.toggled.connect(lambda: self._update_unit_mode("dbm"))
        self.rb_volts.toggled.connect(lambda: self._update_unit_mode("volts"))
        
        # Set fixed size policy to make alignment more predictable
        unit_group.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, 
                                QtWidgets.QSizePolicy.Policy.Preferred)
        
        # Add the unit controls to the layout
        toolbar_layout.addWidget(unit_group)
        
    def _setup_zoom_box_control(self, toolbar_layout):
        """Set up the zoom box mode control."""
        zoom_box_cb = QtWidgets.QCheckBox("Zoom Box Mode")
        zoom_box_cb.setChecked(self.zoom_box_mode)
        zoom_box_cb.setToolTip("When enabled, left-click drag creates a zoom box. When disabled, left-click drag pans.")
        zoom_box_cb.toggled.connect(self._toggle_zoom_box)
        
        # Store reference to the checkbox
        self.zoom_box_cb = zoom_box_cb
        
        toolbar_layout.addWidget(zoom_box_cb)

    def _setup_progress_bars(self, layout):
        """Set up progress bars for each module."""
        self.progress_group = None
        if self.modules:
            self.progress_group = QtWidgets.QGroupBox("Analysis Progress")
            progress_layout = QtWidgets.QVBoxLayout(self.progress_group)
            
            self.progress_bars = {}
            self.progress_labels = {}  # Add labels to show amplitude progress
            for module in self.modules:
                vlayout = QtWidgets.QVBoxLayout()
                
                # Main progress layout
                hlayout = QtWidgets.QHBoxLayout()
                label = QtWidgets.QLabel(f"Module {module}:")
                pbar = QtWidgets.QProgressBar()
                pbar.setRange(0, 100)
                pbar.setValue(0)
                hlayout.addWidget(label)
                hlayout.addWidget(pbar)
                vlayout.addLayout(hlayout)
                
                # Amplitude progress label
                amp_label = QtWidgets.QLabel("")
                amp_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
                vlayout.addWidget(amp_label)
                
                progress_layout.addLayout(vlayout)
                self.progress_bars[module] = pbar
                self.progress_labels[module] = amp_label
                
            layout.addWidget(self.progress_group)
        else:
            self.progress_bars = {}
            self.progress_labels = {}

    def _hide_progress_bars(self):
        """Hide the entire Analysis Progress group."""
        if self.progress_group:
            self.progress_group.hide()

    def _show_progress_bars(self, reset=False):
        """Show the Analysis Progress group again.
           If reset=True, reset progress bars and labels to defaults.
        """
        if self.progress_group:
            self.progress_group.show()
            if reset:
                for module, pbar in self.progress_bars.items():
                    pbar.setValue(0)  # reset progress
                for module, label in self.progress_labels.items():
                    label.clear()     # clear amplitude text
    
    def _setup_plot_area(self, layout):
        """Set up the plot area with tabs for each module."""
        self.tabs = QtWidgets.QTabWidget()
        self.tabs.currentChanged.connect(self._on_active_module_changed) # For Requirement 2
        layout.addWidget(self.tabs)
        
        self.plots = {}
        # Initialize amp_plot and phase_plot to None or a default PlotWidget
        # to ensure they are bound before setXLink is called.
        last_amp_plot: Optional[pg.PlotWidget] = None
        last_phase_plot: Optional[pg.PlotWidget] = None

        for module in self.modules:
            tab = QtWidgets.QWidget()
            tab_layout = QtWidgets.QVBoxLayout(tab)

            # Create amplitude and phase plots with ClickableViewBox
            vb_amp = ClickableViewBox()
            vb_amp.parent_window = self
            vb_amp.module_id = module
            vb_amp.plot_role = 'amp'
            amp_plot = pg.PlotWidget(viewBox=vb_amp, title=f"Module {module} - Magnitude")
            plot_item_amp = amp_plot.getPlotItem()
            if plot_item_amp:
                self._update_amplitude_labels(amp_plot) # amp_plot is PlotWidget, _update_amplitude_labels expects PlotWidget
                plot_item_amp.setLabel('bottom', 'Frequency', units='Hz')
                plot_item_amp.showGrid(x=True, y=True, alpha=0.3)

            vb_phase = ClickableViewBox()
            vb_phase.parent_window = self
            vb_phase.module_id = module
            vb_phase.plot_role = 'phase'
            phase_plot = pg.PlotWidget(viewBox=vb_phase, title=f"Module {module} - Phase")
            plot_item_phase = phase_plot.getPlotItem()
            if plot_item_phase:
                plot_item_phase.setLabel('left', 'Phase', units='deg')
                plot_item_phase.setLabel('bottom', 'Frequency', units='Hz')
                plot_item_phase.showGrid(x=True, y=True, alpha=0.3)
            
            # Add legends for multiple amplitude plots with proper text color
            bg_color, pen_color = ("k", "w") if self.dark_mode else ("w", "k")
            amp_legend = plot_item_amp.addLegend(offset=(30, 10), labelTextColor=pen_color) if plot_item_amp else None
            phase_legend = plot_item_phase.addLegend(offset=(30, 10), labelTextColor=pen_color) if plot_item_phase else None

            # Create curves with periscope color scheme - but don't add data yet
            amp_curve = plot_item_amp.plot([], [], pen=pg.mkPen(TABLEAU10_COLORS[1], width=LINE_WIDTH)) if plot_item_amp else None
            phase_curve = plot_item_phase.plot([], [], pen=pg.mkPen(TABLEAU10_COLORS[0], width=LINE_WIDTH)) if plot_item_phase else None

            tab_layout.addWidget(amp_plot)
            tab_layout.addWidget(phase_plot)
            self.tabs.addTab(tab, f"Module {module}")
            
            self.plots[module] = {
                'amp_plot': amp_plot, # amp_plot is PlotWidget here
                'phase_plot': phase_plot, # phase_plot is PlotWidget here
                'amp_curve': amp_curve,
                'phase_curve': phase_curve,
                'amp_legend': amp_legend,
                'phase_legend': phase_legend,
                'amp_curves': {},  # Will store multiple curves for different amplitudes
                'phase_curves': {},  # Will store multiple curves for different amplitudes
                'resonance_lines_mag': [], # For storing magnitude resonance lines
                'resonance_lines_phase': [] # For storing phase resonance lines
            }
            last_amp_plot = amp_plot
            last_phase_plot = phase_plot
            
        # Apply zoom box mode
        self._apply_zoom_box_mode()

        # Link the x-axis of the last created amplitude and phase plots for synchronized zooming
        if last_phase_plot and last_amp_plot:
            last_phase_plot.setXLink(last_amp_plot)
        
        # Apply initial theme based on dark_mode setting
        if last_amp_plot: self._apply_theme_to_plot(last_amp_plot)
        if last_phase_plot: self._apply_theme_to_plot(last_phase_plot)
        
    def _apply_theme_to_plot(self, plot_widget):
        """Apply the current theme to a specific plot widget."""
        bg_color, pen_color = ("k", "w") if self.dark_mode else ("w", "k")
        plot_widget.setBackground(bg_color)
        
        # Update plot title color using a more direct approach
        plot_item = plot_widget.getPlotItem()
        if plot_item:
            # Set the title explicitly with the color parameter
            title_text = plot_item.titleLabel.text if plot_item.titleLabel else ""
            plot_item.setTitle(title_text, color=pen_color)
            
        # Update axes colors
        for axis_name in ("left", "bottom", "right", "top"):
            ax = plot_widget.getPlotItem().getAxis(axis_name)
            if ax:
                ax.setPen(pen_color)
                ax.setTextPen(pen_color)

    def clear_plots(self):
        """Clear all plots, curves, and legends."""
        for module_id_iter in self.plots: 
            plot_info = self.plots[module_id_iter]
            amp_plot = plot_info['amp_plot']
            phase_plot = plot_info['phase_plot']
            
            plot_info['amp_legend'].clear()
            plot_info['phase_legend'].clear()
            
            for amp, curve in list(plot_info['amp_curves'].items()):
                amp_plot.removeItem(curve)
            for amp, curve in list(plot_info['phase_curves'].items()):
                phase_plot.removeItem(curve)
            
            plot_info['amp_curves'].clear()
            plot_info['phase_curves'].clear()
            
            plot_info['amp_curve'].setData([], [])
            plot_info['phase_curve'].setData([], [])

            self._remove_faux_resonance_legend_entry(module_id_iter)
            self._update_multisweep_button_state(module_id_iter) 


    def update_amplitude_progress(self, module: int, current_amp: int, total_amps: int, amplitude: float):
        """Update the amplitude progress display for a module."""
        if hasattr(self, 'progress_labels') and module in self.progress_labels:
            self.progress_labels[module].setText(f"Amplitude {current_amp}/{total_amps} ({amplitude})")

    def _toggle_normalization(self, checked):
        """Toggle normalization of magnitude plots."""
        self.normalize_magnitudes = checked
        
        for module in self.plots:
            self._update_amplitude_labels(self.plots[module]['amp_plot'])
        
        for module in self.plots:
            if len(self.plots[module]['amp_curves']) > 0:
                self.plots[module]['amp_curve'].setData([], [])
                self.plots[module]['phase_curve'].setData([], [])
        
        for module_id in self.plots: 
            self._update_amplitude_labels(self.plots[module_id]['amp_plot'])

            if module_id in self.raw_data:
                plot_info = self.plots[module_id]
                
                for amp_key, raw_entry_tuple in self.raw_data[module_id].items():
                    amplitude_val, freqs, amps_raw, _, iq_data_raw = self._extract_data_from_tuple(amp_key, raw_entry_tuple)
                    
                    if freqs is None or amps_raw is None or iq_data_raw is None:
                        continue

                    converted_amps = UnitConverter.convert_amplitude(
                        amps_raw, iq_data_raw, self.unit_mode, normalize=self.normalize_magnitudes
                    )
                    
                    if amp_key != 'default':
                        if amplitude_val in plot_info['amp_curves']:
                            plot_info['amp_curves'][amplitude_val].setData(freqs, converted_amps)
                    else: 
                        if not plot_info['amp_curves']:
                            plot_info['amp_curve'].setData(freqs, converted_amps)
                
                plot_info['amp_plot'].autoRange() 
        

    def _toggle_zoom_box(self, enable):
        """Toggle zoom box mode for all plots."""
        self.zoom_box_mode = enable
        self._apply_zoom_box_mode()
        
    def _apply_zoom_box_mode(self):
        """Apply the current zoom box mode setting to all plots."""
        for module in self.plots:
            for plot_type in ['amp_plot', 'phase_plot']:
                viewbox = self.plots[module][plot_type].getViewBox()
                if isinstance(viewbox, ClickableViewBox):
                    viewbox.enableZoomBoxMode(self.zoom_box_mode)

    def _toggle_resonances_visible(self, checked: bool):
        """Show or hide all resonance indicator lines."""
        for module_id in self.plots.keys():
            for line in self.plots[module_id]['resonance_lines_mag']:
                line.setVisible(checked)
            for line in self.plots[module_id]['resonance_lines_phase']:
                line.setVisible(checked)
            self._update_resonance_legend_entry(module_id)

    def _toggle_resonance_edit_mode(self, checked: bool):
        """Enable or disable double-click add/subtract mode."""
        self.add_subtract_mode = checked

    def _update_resonance_checkbox_text(self, module: int):
        """Update the show-resonances checkbox label (count removed)."""
        self.show_resonances_cb.setText("Show Resonances")

    def _add_resonance(self, module: int, freq_hz: float):
        """Add a resonance line at the given frequency."""
        if module not in self.plots:
            return
        plot_info = self.plots[module]
        line_pen = pg.mkPen(RESONANCE_LINE_COLOR, style=QtCore.Qt.PenStyle.DashLine)
        line_mag = pg.InfiniteLine(pos=freq_hz, angle=90, movable=False, pen=line_pen)
        line_phase = pg.InfiniteLine(pos=freq_hz, angle=90, movable=False, pen=line_pen)
        plot_info['amp_plot'].addItem(line_mag)
        plot_info['phase_plot'].addItem(line_phase)
        plot_info['resonance_lines_mag'].append(line_mag)
        plot_info['resonance_lines_phase'].append(line_phase)
        self.resonance_freqs.setdefault(module, []).append(freq_hz)
        self._update_resonance_checkbox_text(module) 
        self._update_resonance_legend_entry(module) 
        self._toggle_resonances_visible(self.show_resonances_cb.isChecked())
        self._update_multisweep_button_state(module)


    def _remove_resonance(self, module: int, freq_hz: float):
        """Remove the nearest resonance line."""
        if module not in self.plots or module not in self.resonance_freqs:
            return
        freqs = self.resonance_freqs[module]
        if not freqs:
            self._update_multisweep_button_state(module) 
            return
        freqs_arr = np.array(freqs)
        idx = int(np.argmin(np.abs(freqs_arr - freq_hz)))
        line_mag = self.plots[module]['resonance_lines_mag'].pop(idx)
        line_phase = self.plots[module]['resonance_lines_phase'].pop(idx)
        self.plots[module]['amp_plot'].removeItem(line_mag)
        self.plots[module]['phase_plot'].removeItem(line_phase)
        freqs.pop(idx)
        self._update_resonance_checkbox_text(module) 
        self._update_resonance_legend_entry(module) 
        self._update_multisweep_button_state(module)

    def _update_unit_mode(self, mode):
        """Update unit mode and redraw only amplitude plots."""
        if mode != self.unit_mode:
            self.unit_mode = mode
            
            for module_id in self.plots: 
                self._update_amplitude_labels(self.plots[module_id]['amp_plot'])

                if module_id in self.raw_data:
                    plot_info = self.plots[module_id]
                    
                    for amp_key, raw_entry_tuple in self.raw_data[module_id].items():
                        amplitude_val, freqs, amps_raw, _, iq_data_raw = self._extract_data_from_tuple(amp_key, raw_entry_tuple)
                        
                        if freqs is None or amps_raw is None or iq_data_raw is None: 
                            continue

                        converted_amps = UnitConverter.convert_amplitude(
                            amps_raw, iq_data_raw, self.unit_mode, normalize=self.normalize_magnitudes
                        )
                        
                        if amp_key != 'default': 
                            if amplitude_val in plot_info['amp_curves']:
                                plot_info['amp_curves'][amplitude_val].setData(freqs, converted_amps)
                        else: 
                            if not plot_info['amp_curves']: 
                                plot_info['amp_curve'].setData(freqs, converted_amps)
                    
                    plot_info['amp_plot'].autoRange() 
            
            self._update_legends_for_unit_mode()

    def _update_legends_for_unit_mode(self):
        """Update the legend entries to reflect the current unit mode."""
        for module in self.plots:
            self.plots[module]['amp_legend'].clear()
            self.plots[module]['phase_legend'].clear()
            
            for amplitude, curve in self.plots[module]['amp_curves'].items():
                label = "" 
                if self.unit_mode == "dbm":
                    if module not in self.dac_scales:
                        print(f"Warning: No DAC scale available for module {module}, cannot display accurate probe power in dBm.")
                        self.rb_counts.setChecked(True)  
                        return  
                    
                    dac_scale = self.dac_scales[module]
                    dbm_value = UnitConverter.normalize_to_dbm(amplitude, dac_scale)
                    label = f"Probe: {dbm_value:.2f} dBm"
                elif self.unit_mode == "volts":
                    if module not in self.dac_scales:
                        print(f"Warning: No DAC scale available for module {module}, cannot display accurate probe amplitude in Volts.")
                        self.rb_counts.setChecked(True)  
                        return  
                    else:
                        dac_scale = self.dac_scales[module]
                        dbm_value = UnitConverter.normalize_to_dbm(amplitude, dac_scale)
                        
                        power_watts = 10**((dbm_value - 30)/10)
                        resistance = 50.0  
                        voltage_rms = np.sqrt(power_watts * resistance)
                        voltage_peak = voltage_rms * np.sqrt(2)
                        voltage_peak = voltage_peak*1e6
                        
                        label = f"Probe: {voltage_peak:.1f} uV (peak)"
                else:  # "counts"
                    label = f"Probe: {amplitude} Normalized Units"
                
                self.plots[module]['amp_legend'].addItem(curve, label)
                
                phase_curve = self.plots[module]['phase_curves'].get(amplitude)
                if phase_curve:
                    self.plots[module]['phase_legend'].addItem(phase_curve, label)         
    
    def _update_amplitude_labels(self, plot):
        """Update plot labels based on current unit mode and normalization state."""
        if self.normalize_magnitudes:
            if self.unit_mode == "dbm":
                plot.setLabel('left', 'Normalized Power', units='dB') 
            else:
                plot.setLabel('left', 'Normalized Magnitude', units='')
        else:
            if self.unit_mode == "counts":
                plot.setLabel('left', 'Magnitude', units='Counts')
            elif self.unit_mode == "dbm":
                plot.setLabel('left', 'Power', units='dBm')
            elif self.unit_mode == "volts":
                plot.setLabel('left', 'Magnitude', units='V')

    def _redraw_all_plots(self):
        """Redraw all plots with current unit mode."""
        for module_id_iter_redraw in self.raw_data: 
            if module_id_iter_redraw in self.plots:
                has_amp_curves = len(self.plots[module_id_iter_redraw]['amp_curves']) > 0
                
                for amp_key, data_tuple in self.raw_data[module_id_iter_redraw].items():
                    if amp_key != 'default':
                        amplitude, freqs, amps, phases, iq_data = self._extract_data_from_tuple(amp_key, data_tuple)
                        
                        if amplitude in self.plots[module_id_iter_redraw]['amp_curves']:
                            converted_amps = UnitConverter.convert_amplitude(
                                amps, iq_data, self.unit_mode, normalize=self.normalize_magnitudes)
                            self.plots[module_id_iter_redraw]['amp_curves'][amplitude].setData(freqs, converted_amps)
                            self.plots[module_id_iter_redraw]['phase_curves'][amplitude].setData(freqs, phases)
                
                if 'default' in self.raw_data[module_id_iter_redraw] and not has_amp_curves:
                    freqs, amps, phases, iq_data = self.raw_data[module_id_iter_redraw]['default']
                    converted_amps = UnitConverter.convert_amplitude(
                        amps, iq_data, self.unit_mode, normalize=self.normalize_magnitudes)
                    self.plots[module_id_iter_redraw]['amp_curve'].setData(freqs, converted_amps)
                    self.plots[module_id_iter_redraw]['phase_curve'].setData(freqs, phases)
                else:
                    self.plots[module_id_iter_redraw]['amp_curve'].setData([], [])
                    self.plots[module_id_iter_redraw]['phase_curve'].setData([], [])
                
                self.plots[module_id_iter_redraw]['amp_plot'].autoRange()
        
        self._update_legends_for_unit_mode()              
    
    def _extract_data_from_tuple(self, amp_key, data_tuple):
        """Extract amplitude and data from a data tuple."""
        if len(data_tuple) == 5:
            freqs, amps, phases, iq_data, amplitude = data_tuple
        else:
            freqs, amps, phases, iq_data = data_tuple
            try:
                amplitude = float(amp_key.split('_')[-1])
            except (ValueError, IndexError):
                amplitude = DEFAULT_AMPLITUDE
        return amplitude, freqs, amps, phases, iq_data

    def _show_find_resonances_dialog(self):
        """Show the dialog to configure and run find_resonances."""
        current_tab_index = self.tabs.currentIndex()
        if current_tab_index < 0:
            QtWidgets.QMessageBox.warning(self, "No Module Selected", "Please select a module tab to analyze.")
            return
        active_module_text = self.tabs.tabText(current_tab_index)
        try:
            active_module = int(active_module_text.split(" ")[1])
        except (IndexError, ValueError):
            QtWidgets.QMessageBox.critical(self, "Error", f"Could not determine active module from tab: {active_module_text}")
            raise
        if not self.raw_data or active_module not in self.raw_data or not self.raw_data[active_module]:
            QtWidgets.QMessageBox.information(self, "No Data", f"No sweep data available for Module {active_module} to find resonances.")
            return

        dialog = FindResonancesDialog(self)
        if dialog.exec():
            params = dialog.get_parameters()
            if params:
                self._run_and_plot_resonances(active_module, params)

    def _run_and_plot_resonances(self, active_module: int, find_resonances_params: dict):
        """Run find_resonances and plot the results on the active module's plots."""
        module_sweeps = self.raw_data.get(active_module)
        if not module_sweeps:
            QtWidgets.QMessageBox.warning(self, "No Data", f"No data found for module {active_module}.")
            self._update_multisweep_button_state(active_module) 
            return

        target_sweep_key = None
        ordered_amplitudes_run = self.original_params.get('amps', [])
        if ordered_amplitudes_run:
            for amp_setting in reversed(ordered_amplitudes_run):
                sweep_key = f"{active_module}_{amp_setting}"
                if sweep_key in module_sweeps:
                    target_sweep_key = sweep_key
                    break
        
        if target_sweep_key is None and 'default' in module_sweeps:
            target_sweep_key = 'default'

        if target_sweep_key is None:
            if module_sweeps:
                target_sweep_key = list(module_sweeps.keys())[-1]


        if target_sweep_key is None:
            QtWidgets.QMessageBox.warning(self, "No Data", f"Could not determine which sweep to analyze for module {active_module}.")
            self._update_multisweep_button_state(active_module) 
            return

        data_tuple = module_sweeps[target_sweep_key]
        
        if len(data_tuple) == 5: 
            frequencies, _, _, iq_complex, _ = data_tuple
        elif len(data_tuple) == 4: 
            frequencies, _, _, iq_complex = data_tuple
        else:
            QtWidgets.QMessageBox.critical(self, "Data Error", "Unexpected data format for the selected sweep.")
            self._update_multisweep_button_state(active_module) 
            return

        if len(frequencies) == 0 or len(iq_complex) == 0:
            QtWidgets.QMessageBox.information(self, "No Data", "Selected sweep has no frequency or IQ data.")
            self._update_multisweep_button_state(active_module) 
            return
            
        try:
            resonance_results = fitting.find_resonances(
                frequencies=frequencies,
                iq_complex=iq_complex,
                module_identifier=f"Module {active_module} (Sweep: {target_sweep_key})",
                **find_resonances_params
            )
        except Exception as e:
            QtWidgets.QMessageBox.critical(self, "Resonance Finding Error", f"Error calling find_resonances: {str(e)}")
            traceback.print_exc()
            self._update_multisweep_button_state(active_module) 
            return

        if active_module in self.plots:
            plot_info = self.plots[active_module]
            amp_plot_item = plot_info['amp_plot'].getPlotItem()
            phase_plot_item = plot_info['phase_plot'].getPlotItem()

            for line in plot_info.get('resonance_lines_mag', []):
                amp_plot_item.removeItem(line)
            plot_info['resonance_lines_mag'] = []

            for line in plot_info.get('resonance_lines_phase', []):
                phase_plot_item.removeItem(line)
            plot_info['resonance_lines_phase'] = []

            self.resonance_freqs[active_module] = []
            
            res_freqs_hz = resonance_results.get('resonance_frequencies', [])

            if not res_freqs_hz:
                QtWidgets.QMessageBox.information(self, "No Resonances Found", 
                                                  f"No resonances were identified for Module {active_module} with the given parameters.")
                self._update_multisweep_button_state(active_module) 
                return 

            line_pen = pg.mkPen('r', style=QtCore.Qt.PenStyle.DashLine)
            for res_freq_hz in res_freqs_hz:
                line_mag = pg.InfiniteLine(pos=res_freq_hz, angle=90, movable=False, pen=line_pen)
                amp_plot_item.addItem(line_mag)
                plot_info['resonance_lines_mag'].append(line_mag)

                line_phase = pg.InfiniteLine(pos=res_freq_hz, angle=90, movable=False, pen=line_pen)
                phase_plot_item.addItem(line_phase)
                plot_info['resonance_lines_phase'].append(line_phase)

            self.resonance_freqs[active_module] = res_freqs_hz
            self._update_resonance_checkbox_text(active_module) 
            self._update_resonance_legend_entry(active_module) 
            self._toggle_resonances_visible(self.show_resonances_cb.isChecked()) 
        self._update_multisweep_button_state(active_module)
        
        # Emit data_ready signal for session auto-export after finding resonances
        # Include filename override if this panel has exported before (to update existing file)
        export_data = self.build_export_dict()  # Use inherited method from mixin
        if self._last_export_filename:
            export_data['_filename_override'] = self._last_export_filename
        self.data_ready.emit(export_data)
        
    def _use_loaded_resonances(self, active_module: int, load_resonance_freqs: list):
        if active_module in self.plots:
            plot_info = self.plots[active_module]
            amp_plot_item = plot_info['amp_plot'].getPlotItem()
            phase_plot_item = plot_info['phase_plot'].getPlotItem()

            for line in plot_info.get('resonance_lines_mag', []):
                amp_plot_item.removeItem(line)
            plot_info['resonance_lines_mag'] = []

            for line in plot_info.get('resonance_lines_phase', []):
                phase_plot_item.removeItem(line)
            plot_info['resonance_lines_phase'] = []

            self.resonance_freqs[active_module] = []
            
            res_freqs_hz = load_resonance_freqs

            line_pen = pg.mkPen('r', style=QtCore.Qt.PenStyle.DashLine)
            for res_freq_hz in res_freqs_hz:
                line_mag = pg.InfiniteLine(pos=res_freq_hz, angle=90, movable=False, pen=line_pen)
                amp_plot_item.addItem(line_mag)
                plot_info['resonance_lines_mag'].append(line_mag)

                line_phase = pg.InfiniteLine(pos=res_freq_hz, angle=90, movable=False, pen=line_pen)
                phase_plot_item.addItem(line_phase)
                plot_info['resonance_lines_phase'].append(line_phase)

            self.resonance_freqs[active_module] = res_freqs_hz
            self._update_resonance_checkbox_text(active_module) 
            self._update_resonance_legend_entry(active_module) 
            self._toggle_resonances_visible(self.show_resonances_cb.isChecked())
        self._update_multisweep_button_state(active_module)
        
    def _remove_faux_resonance_legend_entry(self, module_id: int):
        """Removes the faux resonance legend entry for a module."""
        if module_id in self.plots:
            plot_info = self.plots[module_id]
            amp_legend = plot_info['amp_legend']
            phase_legend = plot_info['phase_legend']

            if module_id in self.faux_resonance_legend_items_mag:
                try:
                    amp_legend.removeItem(self.faux_resonance_legend_items_mag[module_id])
                except Exception: 
                    pass 
                del self.faux_resonance_legend_items_mag[module_id]
            
            if module_id in self.faux_resonance_legend_items_phase:
                try:
                    phase_legend.removeItem(self.faux_resonance_legend_items_phase[module_id])
                except Exception:
                    pass
                del self.faux_resonance_legend_items_phase[module_id]

    def _update_resonance_legend_entry(self, module_id: int):
        """Adds or updates the faux legend entry for resonance count."""
        if module_id not in self.plots:
            return

        self._remove_faux_resonance_legend_entry(module_id) 

        plot_info = self.plots[module_id]
        amp_legend = plot_info['amp_legend']
        phase_legend = plot_info['phase_legend']
        
        count = len(self.resonance_freqs.get(module_id, []))

        if self.show_resonances_cb.isChecked() and count > 0:
            dummy_pen = pg.mkPen('r', style=QtCore.Qt.PenStyle.DashLine)
            
            legend_sample_item_mag = pg.PlotDataItem(pen=dummy_pen)
            amp_legend.addItem(legend_sample_item_mag, f"{count} resonances")
            self.faux_resonance_legend_items_mag[module_id] = legend_sample_item_mag 

            legend_sample_item_phase = pg.PlotDataItem(pen=dummy_pen)
            phase_legend.addItem(legend_sample_item_phase, f"{count} resonances")
            self.faux_resonance_legend_items_phase[module_id] = legend_sample_item_phase


    def _get_periscope_parent(self):
        """
        Get the Periscope parent instance by walking up the parent hierarchy.
        
        Returns:
            The Periscope instance or None if not found
        """
        return find_parent_with_attr(self, 'netanal_windows')
    
    def _check_all_complete(self):
        """
        Check if all progress bars are at 100% and hide the progress group 
        when all analyses are complete.
        """
        if not self.progress_group:
            return
        
        # Walk up parent hierarchy to find Periscope instance
        # (panel may be wrapped in QDockWidget, so parent() might not be Periscope directly)
        parent = self._get_periscope_parent()
        
        if not parent:
            return
            
        window_id = None
        for w_id, w_data in parent.netanal_windows.items():
            if w_data['window'] == self:
                window_id = w_id
                break
        
        if not window_id:
            return
            
        window_data = parent.netanal_windows[window_id]
        
        no_pending_amplitudes = True
        for module in window_data['amplitude_queues']:
            if window_data['amplitude_queues'][module]:
                no_pending_amplitudes = False
                break
        
        all_complete = all(pbar.value() == 100 for pbar in self.progress_bars.values())
        if all_complete and no_pending_amplitudes:
            self.progress_group.setVisible(False)

    def _edit_parameters(self):
        """Open dialog to edit parameters. Re-runs analysis using per-module cable lengths."""
        params_for_dialog = self.current_params.copy()
        params_for_dialog.pop('module_cable_lengths', None)
        params_for_dialog.pop('cable_length', None) 

        dialog = NetworkAnalysisParamsDialog(self, params_for_dialog)
        if dialog.exec():
            updated_general_params = dialog.get_parameters() 
            
            if updated_general_params:
                # Reset the loaded data flag since we're now generating fresh data
                self.is_loaded_data = False
                
                # Update the dock title to remove "(Loaded)" suffix
                periscope = self._get_periscope_parent()
                if periscope:
                    my_dock = periscope.dock_manager.find_dock_for_widget(self)
                    if my_dock:
                        # Get current title and remove " (Loaded)" if present
                        current_title = my_dock.windowTitle()
                        new_title = current_title.replace(" (Loaded)", "")
                        my_dock.setWindowTitle(new_title)
                
                params_for_rerun = updated_general_params.copy()
                params_for_rerun['module_cable_lengths'] = self.module_cable_lengths.copy()
                params_for_rerun.pop('cable_length', None)
                self.current_params = params_for_rerun.copy()
                
                if self.progress_group:
                    self.progress_group.setVisible(True)
                
                parent_widget = self._get_periscope_parent()
                if parent_widget and hasattr(parent_widget, '_rerun_network_analysis'):
                    parent_widget._rerun_network_analysis(self.current_params) # type: ignore

    def _rerun_analysis(self):
        """Re-run the analysis with potentially updated parameters."""
        parent_widget = self._get_periscope_parent()
        if parent_widget and hasattr(parent_widget, '_rerun_network_analysis'):
            params = self.current_params.copy()
            params['module_cable_lengths'] = self.module_cable_lengths.copy()
            params.pop('cable_length', None)

            if self.progress_group:
                self.progress_group.setVisible(True)
            parent_widget._rerun_network_analysis(params) # type: ignore
    
    def set_params(self, params):
        """Set parameters for analysis."""
        self.original_params = params.copy()  
        self.current_params = params.copy()   

        default_cable_length_for_all = params.get('cable_length', DEFAULT_CABLE_LENGTH)
        for mod_id in self.modules: 
            self.module_cable_lengths[mod_id] = params.get('module_cable_lengths', {}).get(mod_id, default_cable_length_for_all)

        if self.tabs.count() > 0:
            self._on_active_module_changed(self.tabs.currentIndex())
        elif self.modules: 
            first_module_id = self.modules[0]
            initial_cable_length = self.module_cable_lengths.get(first_module_id, DEFAULT_CABLE_LENGTH)
            self.cable_length_spin.blockSignals(True)
            self.cable_length_spin.setValue(initial_cable_length)
            self.cable_length_spin.blockSignals(False)
        
        if not hasattr(self, 'plots') or not self.plots:
            return
            
        fmin = params.get('fmin', DEFAULT_MIN_FREQ)
        fmax = params.get('fmax', DEFAULT_MAX_FREQ)
        for module in self.plots:
            self.plots[module]['amp_plot'].setXRange(fmin, fmax)
            self.plots[module]['phase_plot'].setXRange(fmin, fmax)
            self.plots[module]['amp_plot'].enableAutoRange(pg.ViewBox.XAxis, False)
            self.plots[module]['phase_plot'].enableAutoRange(pg.ViewBox.XAxis, False)
            self.plots[module]['amp_plot'].enableAutoRange(pg.ViewBox.YAxis, True)
            self.plots[module]['phase_plot'].enableAutoRange(pg.ViewBox.YAxis, True)
    
    def update_data_with_amp(self, module: int, freqs: np.ndarray, amps: np.ndarray, phases: np.ndarray, amplitude: float):
        """Update the plot data for a specific module and amplitude."""
        iq_data = amps * np.exp(1j * np.radians(phases))  
        key = f"{module}_{amplitude}"
        
        if module not in self.raw_data: self.raw_data[module] = {}
        if module not in self.data: self.data[module] = {}
        
        self.raw_data[module][key] = (freqs, amps, phases, iq_data, amplitude)
        self.data[module][key] = (freqs, amps, phases)
        
        if module in self.plots:
            if len(self.plots[module]['amp_curves']) == 0:
                self.plots[module]['amp_curve'].setData([], [])
                self.plots[module]['phase_curve'].setData([], [])
            
            converted_amps = UnitConverter.convert_amplitude(
                amps, iq_data, self.unit_mode, normalize=self.normalize_magnitudes)
            
            amps_list = self.original_params.get('amps', [amplitude])
            if amplitude in amps_list:
                amp_index = amps_list.index(amplitude)
            else:
                # Fallback if amplitude is not in the predefined list (should ideally not happen if params are consistent)
                # Treat as a new, distinct amplitude for color indexing if it's an unexpected one.
                # This might happen if data is updated with an amplitude not in original_params.
                # For simplicity, let's use a modulo of existing curves if this case is hit,
                # or default to the first color if it's the very first one.
                amp_index = len(self.plots[module]['amp_curves']) % len(TABLEAU10_COLORS)

            num_amps_total = len(amps_list)
            color = None

            if num_amps_total <= 5:
                color = TABLEAU10_COLORS[amp_index % len(TABLEAU10_COLORS)]
            else:
                use_cmap = pg.colormap.get(COLORMAP_CHOICES["AMPLITUDE_SWEEP"])
                # Ensure amp_index is valid for normalization if num_amps_total is 1
                normalized_idx = amp_index / max(1, num_amps_total - 1) if num_amps_total > 1 else 0.0
                if self.dark_mode:
                    # For dark mode, map to [0.3, 1.0]
                    map_value = 0.3 + normalized_idx * 0.7
                else:
                    # For light mode, map to [0.0, 0.75]
                    map_value = normalized_idx * 0.75
                color = use_cmap.map(map_value) if use_cmap else TABLEAU10_COLORS[amp_index % len(TABLEAU10_COLORS)] # Fallback
            
            is_new_curve = False
            if amplitude not in self.plots[module]['amp_curves']:
                is_new_curve = True
                self.plots[module]['amp_curves'][amplitude] = self.plots[module]['amp_plot'].plot(
                    pen=pg.mkPen(color, width=LINE_WIDTH), name=f"Amp: {amplitude}")
                self.plots[module]['phase_curves'][amplitude] = self.plots[module]['phase_plot'].plot(
                    pen=pg.mkPen(color, width=LINE_WIDTH), name=f"Amp: {amplitude}")
            
            self.plots[module]['amp_curves'][amplitude].setData(freqs, converted_amps)
            self.plots[module]['phase_curves'][amplitude].setData(freqs, phases)
            
            if is_new_curve: self._update_legends_for_unit_mode()
        self._update_multisweep_button_state(module) 

    def update_data(self, module: int, freqs: np.ndarray, amps: np.ndarray, phases: np.ndarray):
        """Update the plot data for a specific module."""
        iq_data = amps * np.exp(1j * np.radians(phases))  
        
        if module not in self.raw_data: self.raw_data[module] = {}
        if module not in self.data: self.data[module] = {}
        
        self.raw_data[module]['default'] = (freqs, amps, phases, iq_data)
        self.data[module]['default'] = (freqs, amps, phases)
        
        if module in self.plots:
            if len(self.plots[module]['amp_curves']) == 0:
                converted_amps = UnitConverter.convert_amplitude(
                    amps, iq_data, self.unit_mode, normalize=self.normalize_magnitudes)
                
                freq_ghz = freqs # Keep as freqs, units are handled by axis label
                
                self.plots[module]['amp_curve'].setData(freq_ghz, converted_amps)
                self.plots[module]['phase_curve'].setData(freq_ghz, phases)
        self._update_multisweep_button_state(module) 
    
    def update_progress(self, module: int, progress: float):
        """Update the progress bar for a specific module."""
        if module in self.progress_bars:
            self.progress_bars[module].setValue(int(progress))
    
    def complete_analysis(self, module: int):
        """Mark analysis as complete for a module and emit data_ready signal."""
        if module in self.progress_bars:
            self.progress_bars[module].setValue(100)
            self._check_all_complete()
            
        # Emit data_ready signal for session auto-export
        if self._all_modules_complete():
            export_data = self.build_export_dict()  # Use inherited method from mixin
            self.data_ready.emit(export_data)
    
    def _all_modules_complete(self) -> bool:
        """Check if all modules have completed analysis."""
        if not self.progress_bars:
            return False
        return all(pbar.value() == 100 for pbar in self.progress_bars.values())
            
    def apply_theme(self, dark_mode: bool):
        """Apply the dark/light theme to all plots in this window."""
        self.dark_mode = dark_mode
        
        # Apply theme to all plots
        for module in self.plots:
            plot_info = self.plots[module]
            bg_color, pen_color = ("k", "w") if dark_mode else ("w", "k")
            
            # Apply to amplitude plot
            amp_plot = plot_info['amp_plot']
            amp_plot.setBackground(bg_color)
            
            # Update plot title color - use more direct approach
            amp_plot_item = amp_plot.getPlotItem()
            if amp_plot_item and hasattr(amp_plot_item, 'titleLabel'):
                title_text = amp_plot_item.titleLabel.text if amp_plot_item.titleLabel.text else f"Module {module} - Magnitude"
                amp_plot_item.setTitle(title_text, color=pen_color)
                
            # Update axes colors
            for axis_name in ("left", "bottom", "right", "top"):
                ax = amp_plot_item.getAxis(axis_name) if amp_plot_item else None
                if ax: 
                    ax.setPen(pen_color)
                    ax.setTextPen(pen_color)
            
            # Update legend text color for amplitude plot using the proper API
            amp_legend = plot_info['amp_legend']
            if amp_legend:
                try:
                    amp_legend.setLabelTextColor(pen_color)
                    amp_legend.update()
                except Exception as e:
                    print(f"Error updating amp_legend colors: {e}")
            
            # Apply to phase plot
            phase_plot = plot_info['phase_plot']
            phase_plot.setBackground(bg_color)
            
            # Update plot title color - use more direct approach
            phase_plot_item = phase_plot.getPlotItem()
            if phase_plot_item and hasattr(phase_plot_item, 'titleLabel'):
                title_text = phase_plot_item.titleLabel.text if phase_plot_item.titleLabel.text else f"Module {module} - Phase"
                phase_plot_item.setTitle(title_text, color=pen_color)
                
            # Update axes colors
            for axis_name in ("left", "bottom", "right", "top"):
                ax = phase_plot_item.getAxis(axis_name) if phase_plot_item else None
                if ax:
                    ax.setPen(pen_color)
                    ax.setTextPen(pen_color)
            
            # Update legend text color for phase plot using the proper API
            phase_legend = plot_info['phase_legend']
            if phase_legend:
                try:
                    phase_legend.setLabelTextColor(pen_color)
                    # No need to call phase_legend.update() here, setLabelTextColor should suffice
                except Exception as e:
                    print(f"Error updating phase_legend colors: {e}")
            
            # Redraw plots to ensure all legend items are updated correctly
            self._redraw_all_plots()
            
            # Update resonance legend entry to apply new theme colors
            self._update_resonance_legend_entry(module)

    def _handle_set_cable_length_error(self, module_id: int, error_message: str):
        """Handles error during setting of cable length."""
        QtWidgets.QMessageBox.warning(self, "Set Cable Length Error", 
                                      f"Module {module_id}: {error_message}")



"""Network analysis parameter dialogs."""

from .utils import (
    QtWidgets, QtCore, QRegularExpression, QRegularExpressionValidator,
    QDoubleValidator, QIntValidator,
    DEFAULT_AMPLITUDE, DEFAULT_MIN_FREQ, DEFAULT_MAX_FREQ, DEFAULT_CABLE_LENGTH,
    DEFAULT_NPOINTS, DEFAULT_NSAMPLES, DEFAULT_MAX_CHANNELS, DEFAULT_MAX_SPAN,
    UnitConverter, traceback
)
import pickle
from .tasks import DACScaleFetcher
from .network_analysis_base import NetworkAnalysisDialogBase


def load_network_analysis_payload(parent: QtWidgets.QWidget, file_path: str | None = None):

    # If no filename was passed, fall back to showing a dialog (blocking version)
    if file_path is None:
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.Option.DontUseNativeDialog
        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent,
            "Load Network Analysis Parameters",
            "",
            "Pickle Files (*.pkl *.pickle);;All Files (*)",
            options=options,
        )

    if not file_path:
        return None

    try:
        with open(file_path, "rb") as fh:
            payload = pickle.load(fh)
    except Exception as exc:
        QtWidgets.QMessageBox.critical(
            parent,
            "Load Failed",
            f"Could not read '{file_path}':\n{exc}",
        )
        return None

    if (
        isinstance(payload, dict)
        and isinstance(payload.get("parameters"), dict)
        and isinstance(payload.get("modules"), dict)
    ):
        return payload

    QtWidgets.QMessageBox.warning(
        parent,
        "Invalid File",
        "The selected file does not contain network-analysis parameters.",
    )
    return None



class NetworkAnalysisDialog(NetworkAnalysisDialogBase):
    """
    Dialog for configuring and initiating a new Network Analysis.
    It allows users to specify parameters like frequency range, amplitude,
    modules to scan, and other analysis settings.
    """
    def __init__(self, parent: QtWidgets.QWidget = None, modules: list[int] = None,
                 dac_scales: dict[int, float] = None):
        """
        Initializes the Network Analysis configuration dialog.

        Args:
            parent: The parent widget.
            modules: List of available module numbers.
            dac_scales: Pre-fetched DAC scales for the modules.
        """
        super().__init__(parent, params=None, modules=modules, dac_scales=dac_scales)
        self.setWindowTitle("Network Analysis Configuration")
        self.setModal(False) # Modeless dialog
        self._setup_ui()
        self.load_data_available = False
        self._load_data = {}
        
    def _setup_ui(self):
        """Sets up the user interface elements for the dialog."""
        layout = QtWidgets.QVBoxLayout(self)

        self.import_button = QtWidgets.QPushButton("Import Data")
        self.import_button.clicked.connect(self._load_netanal_data)
        layout.addWidget(self.import_button, alignment=QtCore.Qt.AlignLeft)
        
        param_group = QtWidgets.QGroupBox("Analysis Parameters")
        param_layout = QtWidgets.QFormLayout(param_group)
        
        self.module_entry = QtWidgets.QLineEdit("All")
        self.module_entry.setToolTip("Specify modules to analyze (e.g., '1,2,5', '1-4', or 'All' for modules 1-8).")
        self.module_entry.textChanged.connect(self._update_dac_scale_info) # Update DAC info when module selection changes
        param_layout.addRow("Modules:", self.module_entry)
        
        self.fmin_edit = QtWidgets.QLineEdit(str(DEFAULT_MIN_FREQ / 1e6)) # Default in MHz
        self.fmax_edit = QtWidgets.QLineEdit(str(DEFAULT_MAX_FREQ / 1e6)) # Default in MHz
        param_layout.addRow("Min Frequency (MHz):", self.fmin_edit)
        param_layout.addRow("Max Frequency (MHz):", self.fmax_edit)

        self.cable_length_edit = QtWidgets.QLineEdit(str(DEFAULT_CABLE_LENGTH)) # Default cable length
        param_layout.addRow("Cable Length (m):", self.cable_length_edit)
        
        self.setup_amplitude_group(param_layout) # Add shared amplitude settings
        
        self.points_edit = QtWidgets.QLineEdit(str(DEFAULT_NPOINTS))
        param_layout.addRow("Number of Points:", self.points_edit)
        
        self.samples_edit = QtWidgets.QLineEdit(str(DEFAULT_NSAMPLES))
        param_layout.addRow("Samples to Average:", self.samples_edit)
        
        self.max_chans_edit = QtWidgets.QLineEdit(str(DEFAULT_MAX_CHANNELS))
        param_layout.addRow("Max Channels:", self.max_chans_edit)
        
        self.max_span_edit = QtWidgets.QLineEdit(str(DEFAULT_MAX_SPAN / 1e6)) # Default in MHz
        param_layout.addRow("Max Span (MHz):", self.max_span_edit)
        
        self.clear_channels_cb = QtWidgets.QCheckBox("Clear all channels first")
        self.clear_channels_cb.setChecked(True) # Default to clearing channels
        param_layout.addRow("", self.clear_channels_cb)
        
        layout.addWidget(param_group)
        
        # Buttons for starting or canceling the analysis
        btn_layout = QtWidgets.QHBoxLayout()
        self.start_btn = QtWidgets.QPushButton("Start Analysis")
        self.load_btn = QtWidgets.QPushButton("Load Analysis")
        self.load_btn.setEnabled(False) ### Will enable once file is available.
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        btn_layout.addWidget(self.start_btn)
        btn_layout.addWidget(self.load_btn)
        btn_layout.addWidget(self.cancel_btn)
        layout.addLayout(btn_layout)
        
        self.start_btn.clicked.connect(self.accept) # Connect to QDialog's accept slot

        self.load_btn.clicked.connect(self._load_data_avail) 
        
        self.cancel_btn.clicked.connect(self.reject) # Connect to QDialog's reject slot
        
        self._update_dbm_from_normalized() # Initial update of dBm field based on default amplitude
        self.setMinimumSize(500, 600) # Set a reasonable minimum size
        
    def _get_selected_modules(self) -> list[int]:
        """
        Parses the module entry text to determine the list of selected modules.
        Supports "All", comma-separated values, and ranges (e.g., "1-4").

        Returns:
            A list of integer module IDs. Returns modules 1-8 if "All" is specified.
        """
        module_text = self.module_entry.text().strip()
        selected_modules = []
        if module_text.lower() == 'all':
            selected_modules = list(range(1, 9)) # "All" implies modules 1 through 8
        else:
            for part in module_text.split(','):
                part = part.strip()
                if '-' in part: # Handle ranges like "1-4"
                    try:
                        start, end = map(int, part.split('-'))
                        if start <= end: # Ensure valid range
                            selected_modules.extend(range(start, end + 1))
                    except ValueError:
                        continue # Skip malformed range parts
                elif part: # Handle single numbers
                    try:
                        selected_modules.append(int(part))
                    except ValueError:
                        continue # Skip non-integer parts
        # Remove duplicates and sort, though current logic might not produce duplicates if input is clean.
        return sorted(list(set(selected_modules)))

    
    def _load_data_avail(self):
        ''' Use loaded file parameters, accept the dialog'''
        self.load_data_available = True
        self.accept()


    def _load_netanal_data(self):
        """
        Asynchronously trigger a non-blocking QFileDialog using the main thread.
        Using QTimer avoids interfering with existing threads.
        """
        QtCore.QTimer.singleShot(0, self._open_file_dialog_async)


    def _open_file_dialog_async(self):
        """Open a non-blocking file dialog for selecting a Network Analysis parameter file."""
        if not hasattr(self, "_file_dialog") or self._file_dialog is None:
            self._file_dialog = QtWidgets.QFileDialog(self, "Load Network Analysis Parameters")
            self._file_dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFile)
    
            self._file_dialog.setNameFilters([
                "Pickle Files (*.pkl *.pickle)",
                "All Files (*)",
            ])
    
            # Force non-native + non-blocking behavior
            self._file_dialog.setOptions(
                QtWidgets.QFileDialog.Option.DontUseNativeDialog
                | QtWidgets.QFileDialog.Option.ReadOnly
            )
            self._file_dialog.setModal(False)
    
            # Connect signals
            self._file_dialog.fileSelected.connect(self._on_file_selected)
            self._file_dialog.rejected.connect(self._on_file_dialog_closed)
    
        # Show the dialog async
        self._file_dialog.open()


    @QtCore.pyqtSlot(str)
    def _on_file_selected(self, path: str):
        """Handle file selection, load bias data, and populate the UI fields with file contents."""
        payload = load_network_analysis_payload(self, file_path=path)
        if payload is None:
            return
    
        self.load_btn.setEnabled(True)
        self._load_data = payload.copy()
    
        params = payload["parameters"]
    
        module_value = params.get("module")
        if module_value is None:
            module_text = "All"
        elif isinstance(module_value, (list, tuple, set)):
            module_text = ", ".join(str(int(m)) for m in module_value)
        else:
            module_text = str(module_value)
        self.module_entry.setText(module_text)
    
        amps = params.get("amps") or ([params["amp"]] if "amp" in params else None)
        if amps:
            try:
                amp_text = ", ".join(f"{float(amp):g}" for amp in amps)
            except (TypeError, ValueError):
                amp_text = ", ".join(str(amp) for amp in amps)
            self.amp_edit.setText(amp_text)
    
        def set_if_present(key, widget, formatter):
            if key in params and params[key] is not None:
                try:
                    widget.setText(formatter(params[key]))
                except Exception:
                    widget.setText(str(params[key]))
    
        set_if_present("fmin", self.fmin_edit, lambda v: f"{float(v) / 1e6:g}")
        set_if_present("fmax", self.fmax_edit, lambda v: f"{float(v) / 1e6:g}")
        set_if_present("cable_length", self.cable_length_edit, lambda v: f"{float(v):g}")
        set_if_present("npoints", self.points_edit, lambda v: str(int(float(v))))
        set_if_present("nsamps", self.samples_edit, lambda v: str(int(float(v))))
        set_if_present("max_chans", self.max_chans_edit, lambda v: str(int(float(v))))
        set_if_present("max_span", self.max_span_edit, lambda v: f"{float(v) / 1e6:g}")
    
        if "clear_channels" in params:
            self.clear_channels_cb.setChecked(bool(params["clear_channels"]))
    
        self._update_dac_scale_info()
        self._update_dbm_from_normalized()


    @QtCore.pyqtSlot()
    def _on_file_dialog_closed(self):
        """Handle the event when the file dialog is closed without selection."""
        pass
        
    def get_parameters(self) -> dict | None:
        """
        Retrieves and validates the network analysis parameters from the UI fields.

        Returns:
            A dictionary of parameters if valid, otherwise None.
            Shows an error message on invalid input.
        """
        try:
            if self.load_data_available:
                return self._load_data
            else:
                module_text = self.module_entry.text().strip()
                selected_module_param = None # Parameter for 'module' key
                if module_text.lower() != 'all':
                    parsed_modules = self._get_selected_modules()
                    if parsed_modules:
                        selected_module_param = parsed_modules
                
                amp_text = self.amp_edit.text().strip()
                # Use parsed amplitude values, or default if input is empty
                amps_list = self._parse_amplitude_values(amp_text) or [DEFAULT_AMPLITUDE]
                
                # Construct parameters dictionary
                # Using eval for frequency and span allows expressions, but ensure inputs are numbers.
                # Consider replacing eval with direct float conversion if expressions aren't strictly needed.
                params_dict = {
                    'amps': amps_list,
                    'module': selected_module_param, # Can be None (interpreted as all by backend) or list of modules
                    'fmin': float(eval(self.fmin_edit.text())) * 1e6,  # Convert MHz to Hz
                    'fmax': float(eval(self.fmax_edit.text())) * 1e6,  # Convert MHz to Hz
                    'cable_length': float(self.cable_length_edit.text()),
                    'npoints': int(self.points_edit.text()),
                    'nsamps': int(self.samples_edit.text()),
                    'max_chans': int(self.max_chans_edit.text()),
                    'max_span': float(eval(self.max_span_edit.text())) * 1e6, # Convert MHz to Hz
                    'clear_channels': self.clear_channels_cb.isChecked()
                }
                # Basic validation for frequency range
                if params_dict['fmin'] >= params_dict['fmax']:
                    QtWidgets.QMessageBox.warning(self, "Input Error", "Min Frequency must be less than Max Frequency.")
                    return None
                return params_dict
        except Exception as e:
            traceback.print_exc() # Log the full traceback for debugging
            QtWidgets.QMessageBox.critical(self, "Error Parsing Parameters", f"Invalid parameter input: {str(e)}")
            return None


class NetworkAnalysisParamsDialog(NetworkAnalysisDialogBase):
    """
    Dialog for editing existing Network Analysis parameters.
    This dialog is typically modal and pre-filled with current analysis parameters.
    It fetches DAC scales asynchronously if a CRS object is available from its parent.
    """
    def __init__(self, parent: QtWidgets.QWidget = None, params: dict = None):
        """
        Initializes the dialog for editing network analysis parameters.

        Args:
            parent: The parent widget.
            params: Dictionary of existing parameters to populate the fields.
        """
        super().__init__(parent, params=params) # Pass params to base class
        self.setWindowTitle("Edit Network Analysis Parameters")
        self.setModal(True) # Modal dialog
        self._setup_ui()

        # Attempt to fetch DAC scales if CRS is available from the main window hierarchy
        if parent and hasattr(parent, 'parent') and parent.parent() is not None:
            # Assuming parent.parent() is the main Periscope window
            main_periscope_window = parent.parent() 
            if hasattr(main_periscope_window, 'crs') and main_periscope_window.crs is not None:
                self._fetch_dac_scales(main_periscope_window.crs)
        
    def _fetch_dac_scales(self, crs_obj):
        """
        Initiates asynchronous fetching of DAC scales using DACScaleFetcher.

        Args:
            crs_obj: The CRS (Control and Readout System) object to query for scales.
        """
        self.fetcher = DACScaleFetcher(crs_obj)
        self.fetcher.dac_scales_ready.connect(self._on_dac_scales_ready)
        self.fetcher.start() # Start the QThread for fetching
    
    @QtCore.pyqtSlot(dict)
    def _on_dac_scales_ready(self, scales_dict: dict[int, float]):
        """
        Slot to handle the reception of fetched DAC scales.
        Updates the internal DAC scales and refreshes relevant UI elements.

        Args:
            scales_dict: Dictionary mapping module ID to DAC scale (dBm).
        """
        self.dac_scales = scales_dict
        self._update_dac_scale_info() # Update the DAC scale display label
        self._update_dbm_from_normalized() # Recalculate dBm based on new scales
    
    def _setup_ui(self):
        """Sets up the user interface elements for the dialog."""
        layout = QtWidgets.QVBoxLayout(self)
        form = QtWidgets.QFormLayout()
        
        # Populate fields with existing parameters or defaults
        fmin_mhz = str(self.params.get('fmin', DEFAULT_MIN_FREQ) / 1e6)
        fmax_mhz = str(self.params.get('fmax', DEFAULT_MAX_FREQ) / 1e6)
        self.fmin_edit = QtWidgets.QLineEdit(fmin_mhz)
        self.fmax_edit = QtWidgets.QLineEdit(fmax_mhz)
        form.addRow("Min Frequency (MHz):", self.fmin_edit)
        form.addRow("Max Frequency (MHz):", self.fmax_edit)
        
        self.setup_amplitude_group(form) # Add shared amplitude settings
        
        self.points_edit = QtWidgets.QLineEdit(str(self.params.get('npoints', DEFAULT_NPOINTS)))
        form.addRow("Number of Points:", self.points_edit)
        
        self.samples_edit = QtWidgets.QLineEdit(str(self.params.get('nsamps', DEFAULT_NSAMPLES)))
        form.addRow("Samples to Average:", self.samples_edit)
        
        self.max_chans_edit = QtWidgets.QLineEdit(str(self.params.get('max_chans', DEFAULT_MAX_CHANNELS)))
        form.addRow("Max Channels:", self.max_chans_edit)
        
        max_span_mhz = str(self.params.get('max_span', DEFAULT_MAX_SPAN) / 1e6)
        self.max_span_edit = QtWidgets.QLineEdit(max_span_mhz)
        form.addRow("Max Span (MHz):", self.max_span_edit)
        
        self.clear_channels_cb = QtWidgets.QCheckBox("Clear all channels first")
        self.clear_channels_cb.setChecked(self.params.get('clear_channels', True))
        form.addRow("", self.clear_channels_cb)
        
        layout.addLayout(form)
        
        # OK and Cancel buttons
        btn_layout = QtWidgets.QHBoxLayout()
        self.ok_btn = QtWidgets.QPushButton("OK")
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        btn_layout.addWidget(self.ok_btn)
        btn_layout.addWidget(self.cancel_btn)
        layout.addLayout(btn_layout)
        
        self.ok_btn.clicked.connect(self.accept)
        self.cancel_btn.clicked.connect(self.reject)
        
        # Initial update of dBm field, especially if DAC scales were passed in constructor
        # or if _fetch_dac_scales is not called (e.g., no CRS object).
        self._update_dac_scale_info() # Call this first to set up dac_scale_info label correctly
        self._update_dbm_from_normalized() 
        self.setMinimumSize(500, 600)

    def _get_selected_modules(self) -> list[int]:
        """
        Determines the modules relevant for DAC scale display in this dialog.
        Uses the 'module' parameter passed during initialization. If 'module'
        is None or not specified, it defaults to all modules (1-8).

        Returns:
            A list of integer module IDs.
        """
        selected_module_param = self.params.get('module')
        if selected_module_param is None:
            # If no specific module(s) defined in params, assume all for DAC display purposes
            return list(range(1, 9)) 
        # Ensure it's a list, even if a single int was passed in params
        return selected_module_param if isinstance(selected_module_param, list) else [selected_module_param]
    
    def get_parameters(self) -> dict | None:
        """
        Retrieves and validates the edited network analysis parameters.

        Returns:
            A dictionary of parameters if valid, otherwise None.
            Shows an error message on invalid input.
        """
        try:
            amp_text = self.amp_edit.text().strip()
            amps_list = self._parse_amplitude_values(amp_text) or [self.params.get('amp', DEFAULT_AMPLITUDE)]
            
            # Start with a copy of existing params and update with UI values
            params_dict = self.params.copy() 
            params_dict.update({
                'amps': amps_list,
                'amp': amps_list[0] if amps_list else DEFAULT_AMPLITUDE, # Update single 'amp' for compatibility
                'fmin': float(eval(self.fmin_edit.text())) * 1e6,
                'fmax': float(eval(self.fmax_edit.text())) * 1e6,
                'npoints': int(self.points_edit.text()),
                'nsamps': int(self.samples_edit.text()),
                'max_chans': int(self.max_chans_edit.text()),
                'max_span': float(eval(self.max_span_edit.text())) * 1e6,
                'clear_channels': self.clear_channels_cb.isChecked()
            })
            # Basic validation for frequency range
            if params_dict['fmin'] >= params_dict['fmax']:
                QtWidgets.QMessageBox.warning(self, "Input Error", "Min Frequency must be less than Max Frequency.")
                return None
            return params_dict
        except Exception as e:
            QtWidgets.QMessageBox.critical(self, "Error Parsing Parameters", f"Invalid parameter input: {str(e)}")



from importlib import util
from pathlib import Path
import pytest
import asyncio
import numpy as np
from unittest.mock import Mock, AsyncMock, patch, MagicMock
import sys
import os

# point to the repo root if your test lives elsewhere
demo_path = Path(__file__).resolve().parents[1] / "home" / "Demos" / "simplified_tuning_flow.py"

spec = util.spec_from_file_location("simplified_tuning_flow", demo_path)
simplified_tuning_flow = util.module_from_spec(spec)
spec.loader.exec_module(simplified_tuning_flow)

main = simplified_tuning_flow.main
run_algorithm_flow = simplified_tuning_flow.run_algorithm_flow
sys.modules["simplified_tuning_flow"] = simplified_tuning_flow

class TestIntegration:
    """Full integration tests."""
    
    @pytest.mark.asyncio
    @pytest.mark.integration
    async def test_mock_mode_execution(self):
        """Test that mock mode executes without errors."""
        print("\n=== Testing mock mode execution ===")
        exit_code = await main(serial="MOCK")
        assert exit_code == 0, "Mock mode should complete successfully"
        print("Mock mode completed successfully")

class TestSimpleTuningFlow:
    """Test flow for simplified_tuning_flow.py"""

    @pytest.mark.asyncio
    async def test_network_analysis_step(self):
        """Test the network analysis step in isolation."""
        print("\n=== Testing network analysis step ===")
        
        # Create a minimal mock CRS
        mock_crs = AsyncMock()
        mock_crs.serial = "0000"
        
        # Mock the take_netanal method
        frequencies = np.linspace(100e6, 2450e6, 1000)
        iq_complex = np.random.randn(1000) + 1j * np.random.randn(1000)
        phase_degrees = np.random.randn(1000) * 180
        
        mock_crs.take_netanal = AsyncMock(return_value={
            'frequencies': frequencies,
            'iq_complex': iq_complex,
            'phase_degrees': phase_degrees
        })
        mock_crs.get_cable_length = AsyncMock(return_value=0.0)
        mock_crs.set_cable_length = AsyncMock()
        mock_crs.clear_channels = AsyncMock()
        mock_crs.py_get_samples = AsyncMock()
        
        # Mock find_resonances to return no resonances (stops flow early)
        with patch('simplified_tuning_flow.find_resonances') as mock_find:
            mock_find.return_value = {
                'resonance_frequencies': [],
                'resonances_details': []
            }
            
            # Run the flow
            MODULE = 1
            NETANAL_PARAMS = {
                'amp': 0.001,
                'fmin': 100e6,
                'fmax': 2450e6,
                'nsamps': 10,
                'npoints': 1000,
                'max_chans': 1023,
                'max_span': 500e6,
                'module': MODULE
            }
            
            await run_algorithm_flow(
                mock_crs, MODULE, NETANAL_PARAMS, {}, {}, {}, {}, full_run = False
            )
        
        # Verify network analysis was called
        assert mock_crs.take_netanal.called, "take_netanal should be called"
        call_kwargs = mock_crs.take_netanal.call_args[1]
        assert call_kwargs['fmin'] == 100e6
        assert call_kwargs['fmax'] == 2450e6
        print("Network analysis called with correct parameters")

    @pytest.mark.asyncio
    async def test_cable_delay_correction(self):
        """Test that cable delay correction is applied."""
        print("\n=== Testing cable delay correction ===")
        
        mock_crs = AsyncMock()
        mock_crs.serial = "0000"
        
        # Create data with known cable delay
        frequencies = np.linspace(100e6, 2450e6, 100)
        tau = 10e-9  # 10 ns delay
        phase_degrees = -360 * frequencies * tau
        
        mock_crs.take_netanal = AsyncMock(return_value={
            'frequencies': frequencies,
            'iq_complex': np.exp(1j * np.deg2rad(phase_degrees)),
            'phase_degrees': phase_degrees
        })
        mock_crs.get_cable_length = AsyncMock(return_value=0.0)
        mock_crs.set_cable_length = AsyncMock()
        mock_crs.clear_channels = AsyncMock()
        mock_crs.py_get_samples = AsyncMock()
        
        with patch('simplified_tuning_flow.find_resonances') as mock_find:
            mock_find.return_value = {
                'resonance_frequencies': [],
                'resonances_details': []
            }
            
            await run_algorithm_flow(
                mock_crs, 1,
                {'amp': 0.001, 'fmin': 100e6, 'fmax': 2450e6,
                 'nsamps': 10, 'npoints': 100, 'module': 1},
                {}, {}, {}, {}, full_run = False
            )
        
        # Verify cable length was set
        assert mock_crs.set_cable_length.called, "Cable length should be set"
        call_args = mock_crs.set_cable_length.call_args
        new_length = call_args[1]['length']
        print(f"✓ Cable length set to: {new_length:.2f} m")
    
    @pytest.mark.asyncio
    async def test_resonance_finding(self):
        """Test that resonance finding works correctly."""
        print("\n=== Testing resonance finding ===")
        
        mock_crs = AsyncMock()
        mock_crs.serial = "0000"
        
        # Create data with simulated resonances
        frequencies = np.linspace(400e6, 800e6, 10000)
        
        # Add two resonances
        iq_complex = np.ones(10000, dtype=complex)
        for f0 in [500e6, 600e6]:
            Q = 2e4
            response = 1 / (1 + 2j * Q * (frequencies - f0) / f0)
            iq_complex *= response
        
        phase_degrees = np.angle(iq_complex, deg=True)
        
        mock_crs.take_netanal = AsyncMock(return_value={
            'frequencies': frequencies,
            'iq_complex': iq_complex,
            'phase_degrees': phase_degrees
        })
        mock_crs.get_cable_length = AsyncMock(return_value=0.0)
        mock_crs.set_cable_length = AsyncMock()
        mock_crs.clear_channels = AsyncMock()
        mock_crs.py_get_samples = AsyncMock()
        
        # Don't mock find_resonances - let it run
        # Mock everything after finding resonances
        mock_crs.multisweep = AsyncMock(return_value={})
        
        with patch('simplified_tuning_flow.bias_kids') as mock_bias:
            mock_bias.return_value = {}
            
            await run_algorithm_flow(
                mock_crs, 1,
                {'amp': 0.001, 'fmin': 400e6, 'fmax': 800e6,
                 'nsamps': 10, 'npoints': 10000, 'module': 1},
                {'min_dip_depth_db': 0.5, 'min_Q': 1e4, 'max_Q': 1e7,
                 'min_resonance_separation_hz': 50e3, 'data_exponent': 2.0},
                {'span_hz': 500e3, 'npoints_per_sweep': 50, 'amp': 0.001,
                 'nsamps': 10, 'module': 1},
                {'apply_skewed_fit': False, 'apply_nonlinear_fit': False},
                {'num_samples': 1000, 'module': 1}, full_run = False
            )
        
        print("Resonance finding completed")
    
    @pytest.mark.asyncio
    async def test_full_flow_with_resonances(self):
        """Test complete flow when resonances are found."""
        print("\n=== Testing full flow with resonances ===")
        
        mock_crs = AsyncMock()
        mock_crs.serial = "0000"
        
        # Setup network analysis
        frequencies = np.linspace(100e6, 2450e6, 1000)
        mock_crs.take_netanal = AsyncMock(return_value={
            'frequencies': frequencies,
            'iq_complex': np.random.randn(1000) + 1j * np.random.randn(1000),
            'phase_degrees': np.random.randn(1000) * 180
        })
        mock_crs.get_cable_length = AsyncMock(return_value=0.0)
        mock_crs.set_cable_length = AsyncMock()
        mock_crs.clear_channels = AsyncMock()
        
        # Mock resonances
        test_resonances = [500e6, 600e6, 700e6]
        
        # Mock multisweep results
        multisweep_results = {}
        for i, f in enumerate(test_resonances):
            freqs = np.linspace(f - 250e3, f + 250e3, 50)
            multisweep_results[i] = {
                'frequencies': freqs,
                'iq_complex': np.random.randn(50) + 1j * np.random.randn(50),
                'bias_frequency': f,
                'fit_params': {
                    'fr': f,
                    'Qr': 1e4,
                    'Qc': 5e4,
                    'Qi': 5e4
                }
            }
        
        mock_crs.multisweep = AsyncMock(return_value=multisweep_results)
        
        # Mock noise data
        mock_spectrum = Mock()
        mock_spectrum.psd_i = [np.random.randn(100) for _ in range(3)]
        mock_spectrum.psd_q = [np.random.randn(100) for _ in range(3)]
        mock_spectrum.freq_iq = np.linspace(0, 500, 100)
        mock_spectrum.freq_dsb = np.linspace(-500, 500, 100)
        
        mock_samples = Mock()
        mock_samples.spectrum = mock_spectrum
        mock_crs.py_get_samples = AsyncMock(return_value=mock_samples)
        
        # Patch the algorithm functions
        with patch('simplified_tuning_flow.find_resonances') as mock_find, \
             patch('simplified_tuning_flow.bias_kids') as mock_bias, \
             patch('simplified_tuning_flow.fit_skewed_multisweep') as mock_fit:
            
            # Mock find_resonances
            mock_find.return_value = {
                'resonance_frequencies': test_resonances,
                'resonances_details': [
                    {'q_estimated': 1e4, 'prominence_db': 3.0}
                    for _ in test_resonances
                ]
            }
            
            # Mock fitting
            mock_fit.return_value = multisweep_results
            
            # Mock bias
            bias_results = {}
            for i, f in enumerate(test_resonances):
                bias_results[i] = {
                    'bias_frequency': f,
                    'df_calibration': 1e6
                }
            mock_bias.return_value = bias_results
            
            # Run full flow
            await run_algorithm_flow(
                mock_crs, 1,
                {'amp': 0.001, 'fmin': 100e6, 'fmax': 2450e6,
                 'nsamps': 10, 'npoints': 1000, 'module': 1},
                {'min_dip_depth_db': 1.0, 'min_Q': 1e4, 'max_Q': 1e7,
                 'min_resonance_separation_hz': 100e3, 'data_exponent': 2.0},
                {'span_hz': 500e3, 'npoints_per_sweep': 50, 'amp': 0.001,
                 'nsamps': 10, 'module': 1, 'bias_frequency_method': 'max-diq',
                 'rotate_saved_data': False, 'sweep_direction': 'upward'},
                {'apply_skewed_fit': True, 'apply_nonlinear_fit': False,
                 'approx_Q_for_fit': 1e4, 'fit_resonances': True,
                 'center_iq_circle': True, 'normalize_fit': True},
                {'num_samples': 1000, 'return_spectrum': True,
                 'scaling': 'psd', 'reference': 'absolute',
                 'nsegments': 5, 'spectrum_cutoff': 0.9,
                 'channel': None, 'module': 1}, full_run = False
            )
        
        # Verify all major steps were called
        assert mock_crs.take_netanal.called, "Network analysis should run"
        assert mock_find.called, "Resonance finding should run"
        assert mock_crs.multisweep.called, "Multisweep should run"
        assert mock_bias.called, "Bias should run"
        assert mock_crs.py_get_samples.called, "Noise sampling should run"
        
        print("Full flow completed with all steps")
    
    @pytest.mark.asyncio
    async def test_no_resonances_handling(self):
        """Test that flow handles no resonances gracefully."""
        print("\n=== Testing no resonances case ===")
        
        mock_crs = AsyncMock()
        mock_crs.serial = "0000"
        
        # Flat response - no resonances
        frequencies = np.linspace(100e6, 2450e6, 1000)
        mock_crs.take_netanal = AsyncMock(return_value={
            'frequencies': frequencies,
            'iq_complex': np.ones(1000) + 1j * np.zeros(1000),
            'phase_degrees': np.zeros(1000)
        })
        mock_crs.get_cable_length = AsyncMock(return_value=0.0)
        mock_crs.set_cable_length = AsyncMock()
        mock_crs.clear_channels = AsyncMock()
        mock_crs.multisweep = AsyncMock()
        mock_crs.py_get_samples = AsyncMock()
        
        with patch('simplified_tuning_flow.find_resonances') as mock_find:
            mock_find.return_value = {
                'resonance_frequencies': [],
                'resonances_details': []
            }
            
            await run_algorithm_flow(
                mock_crs, 1,
                {'amp': 0.001, 'fmin': 100e6, 'fmax': 2450e6,
                 'nsamps': 10, 'npoints': 1000, 'module': 1},
                {}, {}, {}, {}, full_run = False
            )
        
        # Should not call multisweep when no resonances
        mock_crs.multisweep.assert_not_called()
        print("Flow correctly handles no resonances")
    
    @pytest.mark.asyncio
    async def test_error_handling(self):
        """Test error handling in main function."""
        print("\n=== Testing error handling ===")
        
        with patch('simplified_tuning_flow.create_mock_crs') as mock_create:
            mock_create.side_effect = Exception("Simulated error")
            
            exit_code = await main(serial="MOCK")
            assert exit_code == 1, "Should return error code on exception"
        
        print("Errors handled gracefully")


if __name__ == "__main__":
    # Run tests with verbose output
    pytest.main([__file__, "-v", "-s"])


"""
Dirfile (libgetdata) interface

Where possible, we try to just expose the underlying pygetdata API - it's not
terribly hard to use, and attempting to wrap it in something "better" is a
fairly weak position.
"""


def load_dirfile(name):
    return Dirfile(name)


class Dirfile:
    """
    Augment pygetdata.dirfile with rfmux-specific features

    This is implemented using __getattr__ shenanigans because we can't directly
    inherit from a pygetdata.dirfile() instance - inheriting from a C/C++ base
    class isn't possible.
    """

    def __init__(self, dirfile_path: str):
        import pygetdata

        self._df = pygetdata.dirfile(dirfile_path)

    def __dir__(self):
        # Allow tab-completion on underlying pygetdata.dirfile() instance
        return list(set(self._df.__dir__() + super().__dir__()))

    def __getattr__(self, name: str):
        # Can't use pygetdata.dirfile as a base class, so here we are instead
        return getattr(self._df, name)

    def get_samples(self, serial, module: int, channel: int):
        """
        Friendly API shim to retrieve rfmux sample data from a dirfile.

        This:

        >>> df.get_samples("0024", module=1, channel=1)

        is just a wrapper around the pygetdata API:

        >>> df.getdata("serial_0024.m01_c0001")
        """

        if isinstance(serial, int):
            serial = f"{serial:04i}"
        elif not isinstance(serial, str):
            raise TypeError("Unexpected type for argument 'serial')")

        if not isinstance(module, int) or not isinstance(channel, int):
            raise TypeError("Unexpected integers for module/channel arguments")

        return self.getdata(f"serial_{serial}.m{module:02d}_c{channel:04d}")



import pytest
from types import SimpleNamespace
from unittest.mock import MagicMock

from rfmux.tools.periscope.app import Periscope
from rfmux.tools.periscope.utils import QtWidgets


@pytest.fixture(scope="session")
def qt_app():
    """
    Provide a QApplication for all UI smoke tests.
    Ensures Qt is initialized only once.
    """
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])
    yield app
    app.quit()


@pytest.fixture
def mock_periscope(qt_app):
    """
    Build a minimal Periscope instance for mocked UI tests.
    Equivalent to the old `_build_mock_periscope()`.
    """

    # Allocate object without running Periscope.__init__
    periscope = Periscope.__new__(Periscope)
    QtWidgets.QMainWindow.__init__(periscope)

    # Required minimal CRS mock
    periscope.crs = MagicMock()
    periscope.crs.TIMESTAMP_PORT = SimpleNamespace(
        BACKPLANE="BACKPLANE",
        TEST="TEST",
        SMA="SMA"
    )

    # Basic runtime attributes
    periscope.module = 1
    periscope.pool = MagicMock()
    periscope.pool.start = MagicMock()

    periscope.channel_list = [[1]]
    periscope.dark_mode = False
    periscope.dac_scales = {1: -0.5}
    periscope.raw_data = {1: [1, 2, 3]}
    periscope.resonance_freqs = {1: [90e6, 91e6]}

    periscope.timer = MagicMock()
    periscope.receiver = MagicMock()
    periscope.receiver.stop = MagicMock()
    periscope.receiver.wait = MagicMock()

    # UI bookkeeping
    periscope.netanal_window_count = 0
    periscope.netanal_windows = {}
    periscope.netanal_tasks = {}

    periscope.multisweep_window_count = 0
    periscope.multisweep_windows = {}
    periscope.multisweep_tasks = {}

    # Mock tab widget
    periscope.tabs = MagicMock()
    periscope.tabs.currentIndex.return_value = 0
    periscope.tabs.tabText.return_value = "Module 1"

    # Signals required by smoke test
    periscope.netanal_signals = MagicMock()
    periscope.multisweep_signals = MagicMock()

    dock_manager = MagicMock()

    def _make_dock(widget, title=None, dock_id=None):
        dock = MagicMock()
        dock.widget.return_value = widget
        dock.windowTitle.return_value = title or ""
        return dock

    dock_manager.create_dock.side_effect = _make_dock
    dock_manager.get_dock.return_value = None
    dock_manager.protect_dock = MagicMock()
    dock_manager.remove_dock = MagicMock()
    periscope.dock_manager = dock_manager

    periscope.setDockNestingEnabled = MagicMock()
    periscope.tabifyDockWidget = MagicMock()
    
    for sig_name in (
        "progress",
        "starting_iteration",
        "data_update",
        "completed_iteration",
        "all_completed",
        "error",
        "fitting_progress",
    ):
        setattr(periscope.multisweep_signals, sig_name, MagicMock())

    return periscope


def test_periscope_ui_smoke(mock_periscope):
    """
    Run the mocked UI smoke test using pytest.
    No real dialogs or Qt threads should be launched.
    """

    # The core smoke-test entry point
    mock_periscope.run_ui_mock_smoke_test()


    # Passing = no exceptions
    assert True